import numpy as np
from copy import deepcopy

class DataLoader(object):
    """
        Creates an object that loads data from csv files and normalizes this data if requested.
        
        Args:
            datapath: The path to where the CSV file can be found. The data 
                should be delimited using commas.
            dataframe: Data can be loaded by the user and specified as a dataframe. 
                If a dataframe is specified, the datapath is automatically ignored.
            variable_names: If the CSV file does not contain the variable names,
                the user can specify them here. If this argument is not specified,
                the variable names are read from the first line of the CSV file
                (default = None). 
            normalize: If switch on, the data will be normalized. The user can 
                choose between 'minmax' or 'zscore' normalization (default = False). 
            delimiter: Specify the symbol used to separate data in the dataset (default = ',').
            verbose: Enable verbose mode (default = True).
    """
    
    def __init__(self, datapath=None, dataframe=None, process_categorical=False, 
        variable_names=None, normalize=False, delimiter=',', verbose=True):

        # user specified a dataframe?
        if dataframe is not None:
            
            import pandas as pd

            if not isinstance(dataframe, pd.DataFrame):
                raise TypeError("Please specify a valid dataframe.")

            # valid dataframe specified: proceed with creating pyFUME's data structures

            new_data_frame = deepcopy(dataframe)
            output_name=dataframe.columns[-1]
            
            if process_categorical:
                if verbose: print(" * Processing categorical data.")
                new_data_frame = pd.get_dummies(new_data_frame) 
                new_data_frame = new_data_frame[ [ col for col in new_data_frame.columns if col != output_name ] + [output_name] ]

            else:
                list_categorical = self._check_categorical(dataframe)    
                if verbose: 
                    print(" * Dropping categorical data.")
                    print("   Detected categorical variables:", list_categorical)
                
                for cat in list_categorical:
                    new_data_frame = new_data_frame.drop(cat, axis=1)
            
            # DEBUG
            new_data_frame.to_csv("__new_dataset.csv", index=False)
            
            # Convert to matrix for pyFUME processing and 
            # store the names of the columns as variable names.
            self.data = new_data_frame.to_numpy()           
            self.variable_names = list(new_data_frame.columns)[:-1]
            
            if verbose:
                print(" * Variable names:", self.variable_names)
                print(" * Dataframe %dx%d imported in pyFUME." % self.data.shape)
            

        # pyFUME's conventional behavior
        else:

            if variable_names is None:
                variable_names = np.genfromtxt(datapath, dtype='str', delimiter=delimiter, max_rows=1)
                if verbose ==True:
                    print('The following variable names were detected from the input file: ', variable_names)
                self.output_name=variable_names[-1]
                self.variable_names=variable_names[:-1]
                self.data=np.genfromtxt(datapath, delimiter=delimiter, skip_header=1, filling_values=np.nan)
            else:
                self.variable_names=variable_names
                self.data=np.genfromtxt(datapath, delimiter=delimiter, filling_values=np.nan)
                
        self.dataX=self.data[:,0:-1]
        self.dataY=self.data[:,-1]

        if normalize=='minmax' or normalize=='linear' or normalize==True:
            mini=np.abs(self.dataX).min(axis=0)
            maxi=np.abs(self.dataX).max(axis=0)
            self.normalization_values=zip(self.variable_names, mini, maxi)

            self.dataX = (self.dataX - self.dataX.min(axis=0) / self.dataX.max(axis=0)-self.dataX.min(axis=0))

        elif normalize == 'zscore':
            self.dataX = (self.dataX - self.dataX.mean(axis=0)) / self.dataX.std(axis=0)
            

    def _check_categorical(self, DF):
        cols = DF.columns[:-1] # last one is ALWAYS the output variable
        num_cols = DF._get_numeric_data().columns
        categ_cols = list( set(cols)-set(num_cols))
        return categ_cols
    
    def get_input_data(self):
        return self.dataX
    
    def get_target_data(self):
        return self.dataY
    def get_all_data(self):
        return self.data
    
    def get_variable_names(self):
        return self.variable_names
    
    def get_normalization_values(self):
        return self.normalization_values


if __name__ == "__main__":

    pass


import numpy as np
from scipy.spatial.distance import cdist
from scipy.linalg import norm

#from random import seed

# To do:
# Give option to cluster in input or input-output space?
# More clustering methods
# data input argument


class Clusterer(object):
    """
        Creates a new clusterer object that can cluster the (training) data in the input-output feature space. The user should specify the 'data'argument OR the 'x_train' and 'y_train' argument.
        
        Args:
            nr_clus: Number of clusters that should be identified in the data.  
            x_train: The input data (default = None).
            y_train: The output data (true label/golden standard) (default = None).
            data: The data to be clustered (default = None).
    """ 
    
    def __init__(self, nr_clus, x_train=None, y_train=None, data=None, relational_data=None, verbose=False):
        self.x_train=x_train
        self.y_train=y_train
        self.nr_clus = nr_clus
        self._verbose = verbose

        if data is None and (x_train is None or y_train is None) and relational_data is None:
            raise Exception("Please specify a valid dataset for clustering.")
        elif data is not None:
        	self.data=data
        elif relational_data is not None:
            self.relational_data=relational_data
        else:
            self.data=np.concatenate((self.x_train, np.expand_dims(self.y_train, axis=1)), axis=1)#.reshape(len(self.y_train),1)),axis=1)

                                     
    def cluster(self, method="fcm", **kwargs):
        """
        Clusters the data using the clustering method as specified by the user.
            
        Args:
                method: The method used for the clustering. The user can choose 'fcm' (fuzzy c-means), 'fst-pso' (fst-pso based clustering) and 'gk' (Gustafson-Kessel).
                **kwargs: Additional arguments to change settings of the clustering method.     
                
        Returns: 
                Tuple containing (centers, partion_matrix, jm)   
                    - centers: The location of the identified cluster centers.
                    -  partition_matrix: A matrix containing the cluster memberships of each data point to each of the clusters.
                    - jm: Fitness function of the best solution.
        """ 

        if self._verbose:
            print(" * Clustering method:", method)

        try:
            self.m = kwargs["m"]
        except:
            self.m = 2
            
        if method == "fcm" or method == 'gk' or method == 'GK' or method == 'Gustafson-Kessel' or method == 'gustafson-kessel' or method == 'g-k' or method == "pfcm" or method == "fst-pso" or method == "fstpso":
            assert self.data is not None or (self.x_train is not None and self.y_train is not None), 'Please specify the data tobe clustered.'
        elif method == 'RFCM' or method == "relational_clustering" or method == "relational" or method == "RC":
            assert self.relational_data is not None, 'Please specify the relational data matrix.'
            

        if method == "fcm":
            try:
                max_iter = kwargs["fcm_max_iter"]
            except:
                max_iter = 1000
            try:
                error = kwargs["fcm_error"]
            except:
                error = 0.005

            centers, partition_matrix, jm = self._fcm(data=self.data, n_clusters=self.nr_clus, m=self.m, max_iter=max_iter, error=error)
        
        elif method == 'gk' or method == 'GK' or method == 'Gustafson-Kessel' or method == 'gustafson-kessel' or method == 'g-k':
            try:
                max_iter = kwargs["gk_max_iter"]
            except:
                max_iter=1000
            try:
                error = kwargs["gk_error"]
            except:
                error = 0.005
            centers, partition_matrix, jm =self._gk(m=self.m, max_iter=max_iter) 

        elif method == "pfcm":
            try:
                n = kwargs["pfcm_n"]
            except:
                n = 2
            try:
                max_iter = kwargs["pfcm_max_iter"]
            except:
                max_iter = 1000
            try:
                error = kwargs["pfcm_error"]
            except:
                error = 0.005
            try:
                a = kwargs["pfcm_a"]
            except:
                a = 0.5
            try:
                b = kwargs["pfcm_b"]
            except:
                b = 0.5

            centers, partition_matrix, typicality_matrix, jm = self._pfcm(data=self.data, n_clusters=self.nr_clus, m=self.m, n=n, max_iter=max_iter, error=error, a=a, b=b)    

        elif method == "fst-pso" or method == "fstpso":
            try:
                max_iter = kwargs["fstpso_max_iter"]
            except:
                max_iter=100
            try:
                n_particles = kwargs["fstpso_n_particles"]
            except:
                n_particles = None
            try:
                path_fit_dump = kwargs["fstpso_path_fit_dump"]
            except:
                path_fit_dump = None
            try:
                path_sol_dump = kwargs["fstpso_path_sol_dump"]
            except:
                path_sol_dump = None
            centers, partition_matrix, jm = self._fstpso(data=self.data, n_clusters=self.nr_clus, max_iter=max_iter, n_particles=n_particles, m=self.m, path_fit_dump=path_fit_dump, path_sol_dump=path_sol_dump)
               
        elif method == 'RFCM' or method == 'rfcm' or method == "relational_clustering" or method == "relational" or method == "RC":
            try:
                max_iter = kwargs["RFCM_max_iter"]
            except:
                max_iter=100
            try:
                error = kwargs["RFCM_error"]
            except:
                error = 0.005
            try:
                error = kwargs["RFCM_initialization"]
            except:
                initialization = 'random_initialization'
            
            centers, partition_matrix, jm = self._rfcm(R=self.relational_data, c=self.nr_clus, m = self.m, epsilon = error, maxIter = max_iter, initType = initialization)            
            
        return centers, partition_matrix, jm

    def _fcm(self, data, n_clusters, m=2, max_iter=1000, error=0.005):
        #data: 2d array, size (N, S). N is the number of instances; S is the number of variables
        #n_clusters: number of clusters
        #m: fuzzy clustering coefficient
        #max_it: maximum number of iterations, default=1000
        #error: stopping criterion, default=0.005
        #seed: seed for random initialization of u matrix
        
        n_instances = data.shape[0]
        
        #randomly initaliaze u
        u = np.random.rand(n_instances, n_clusters)
        u = np.fmax(u, np.finfo(np.float64).eps)
        ut = u.T
        
        for it in range(0,max_iter):
            #copy old u matrix
            u_old = ut.copy()
            u_old /= np.ones((n_clusters, 1)).dot(np.atleast_2d(u_old.sum(axis=0)))
            u_old = np.fmax(u_old, np.finfo(np.float64).eps)
        
            #elevate to m
            um = u_old ** m
        
            #calculate cluster centers
            centers = um.dot(data) / (np.ones((data.shape[1], 1)).dot(np.atleast_2d(um.sum(axis=1))).T)
        
            #calculate distances
            dist = cdist(centers, data, metric='euclidean')
            dist = np.fmax(dist, np.finfo(np.float64).eps)
        
            #calculate objective
            jm = (um * dist ** 2).sum()
        
            #calculate new u matrix
            ut = dist ** (- 2. / (m - 1))
            ut /= np.ones((n_clusters, 1)).dot(np.atleast_2d(ut.sum(axis=0)))
        
            #stopping criterion
            if np.linalg.norm(ut - u_old) < error:
                break
        
        partition_matrix = ut.T
        return centers, partition_matrix, jm
    
    def _fuzzy_k_protoypes(self, data, categorical_indices, n_clusters, m=2, max_iter=1000, error=0.005):
        #data: 2d array, size (N, S). N is the number of instances; S is the number of variables
        #n_clusters: number of clusters
        #m: fuzzy clustering coefficient
        #max_it: maximum number of iterations, default=1000
        #error: stopping criterion, default=0.005
        #seed: seed for random initialization of u matrix
        
        n_instances = data.shape[0]
        
        #randomly initaliaze u
        u = np.random.rand(n_instances, n_clusters)
        u = np.fmax(u, np.finfo(np.float64).eps)
        ut = u.T
        
        for it in range(0,max_iter):
            #copy old u matrix
            u_old = ut.copy()
            u_old /= np.ones((n_clusters, 1)).dot(np.atleast_2d(u_old.sum(axis=0)))
            u_old = np.fmax(u_old, np.finfo(np.float64).eps)
        
            #elevate to m
            um = u_old ** m
        
            #calculate cluster centers
            centers = self._FKP_centers(um, data, categorical_indices)
        
            #calculate distances
            dist = self._FKP_distances(centers, data, categorical_indices, numerical_metric='euclidean')
            dist = np.fmax(dist, np.finfo(np.float64).eps)
        
            #calculate objective
            jm = (um * dist ** 2).sum()
        
            #calculate new u matrix
            ut = dist ** (- 2. / (m - 1))
            ut /= np.ones((n_clusters, 1)).dot(np.atleast_2d(ut.sum(axis=0)))
        
            #stopping criterion
            if np.linalg.norm(ut - u_old) < error:
                break
        
        partition_matrix = ut.T
        
        return centers, partition_matrix, jm
    
    def _FKP_centers(self, um, data, categorical_indices):  
        # If data is categorical find the mode
        # Take the max of the sum of the columns of the partition matrix
        
        # if data is numerical, find the mean
        centers = um.dot(data) / (np.ones((data.shape[1], 1)).dot(np.atleast_2d(um.sum(axis=1))).T)
    
        return centers
    
    def _FKP_distances(self, centers, data, categorical_indices, numerical_metric='euclidean'):
        
        # If the data is numerical use euclidean distance
        distances = cdist(centers, data, metric='euclidean')
        
        # If the data is categorical, determine if value is different from the mode
        # if data point == mode, distance = np.finfo(np.float64).eps
        # if data point ~= mode, distance i= 1
        
    
        return distances
    
    def _fstpso(self, data, n_clusters, max_iter=100, n_particles=None, m=2, path_fit_dump=None, path_sol_dump=None):
        #data: 2d array, size (N, S). N is the number of instances; S is the number of variables
        #n_clusters: number of clusters
        #max_iter: number of maximum iterations of FST-PSO, default is 100
        #n_particles: number of particles in the swarm, if None it is automatically set by FST-PSO
        #m: fuzzy clustering coefficient
        #path_fit_dump: path to the file where the best fitness score at each iteration will be dumped
        #path_sol_dump: path to the file where the best solution at each iteration will be dumped

        try:
            from fstpso import FuzzyPSO
        except:
            print("ERROR: please, pip install fst-pso to use this functionality.")

        #n_instances = data.shape[0]
        n_variables = data.shape[1]

        #set search space boundaries
        bounds = [0]*n_variables
        for i in range(n_variables):
            x = min([row[i] for row in data])
            y = max([row[i] for row in data])
            bounds[i] = [x, y]

        search_space = []
        for i in bounds:
            search_space.extend([i]*n_clusters)

        #initializing FST-PSO
        FP = FuzzyPSO()
        FP.set_search_space(search_space)
        if n_particles != None: FP.set_swarm_size(n_particles)

        #generally better results are obtained with this rule disabled
        FP.disable_fuzzyrule_minvelocity()

        #fitness function definition
        def fitness(particle):
            particle = list(map(float,particle))
            centers = np.reshape(particle, (n_variables, n_clusters)).T

            #calculating fitness value of found solution
            dist = cdist(data, centers, metric='sqeuclidean')
            
            um = np.zeros(np.shape(dist))
            for i in range(np.shape(um)[0]):
                for j in range(np.shape(um)[1]):
                    um[i][j] = np.sum(    np.power(    np.divide(    dist[i][j],dist[i])    ,    float(1/(m-1))    )    )
            um = np.reciprocal(um)

            um_power = np.power(um,m)

            fitness_value = np.sum(np.multiply(um_power,dist))
            return fitness_value

        #fitness function setting
        FP.set_fitness(fitness, skip_test=True)

        #execute optimization
        result = FP.solve_with_fstpso(max_iter=max_iter, dump_best_fitness=path_fit_dump, dump_best_solution=path_sol_dump)

        #reshaping centers
        solution = list(map(float,result[0].X))
        centers = np.reshape(solution, (n_variables, n_clusters)).T
        
        #calculating membership matrix
        dist = cdist(data, centers, metric='sqeuclidean')
        um = np.zeros(np.shape(dist))
        for i in range(np.shape(um)[0]):
            for j in range(np.shape(um)[1]):
                um[i][j] = np.sum(    np.power(    np.divide(    dist[i][j],dist[i])    ,    float(1/(m-1))    )    )
        partition_matrix = np.reciprocal(um)

        #final fitness value
        jm =  result[1]

        return centers, partition_matrix, jm
    
    def _pfcm(self, data, n_clusters, m=2, max_iter=1000, error=0.005, a=0.5, b=0.5, n=2):
        
        #data: Dataset to be clustered, with size M-by-N, where M is the number of data points
        #and N is the number of coordinates for each data point.
        #c : Number of clusters
        #m: fuzzy clustering coefficient (default = 2)
        #max_iter: Maximum number of iterations (default = 1000)
        #error : stopping criterion (default=0.005)
        #a: Relative importane of fuzzy membership (default = 0.5)
        #b: Relative importane of typicality values (default = 0.5) 
        #n: User-defined constant n (default = 2)

        #Return values:
        #centers: The locations of the found clusters centers
        #partition_matrix: Partition matrix
        #typicality_matrix: Typicality Matrix
        #jm: The objective funtion for U and T
        

        # Randomly initiaize the partitioning matrix and typicality matrix
        n_instances=len(data)
        partition_matrix = np.random.rand(n_instances, n_clusters)
        partition_matrix = np.fmax(partition_matrix, np.finfo(np.float64).eps)        # avoid 0's in the matrix
        typicality_matrix = np.random.rand(n_instances, n_clusters)
        typicality_matrix = np.fmax(typicality_matrix, np.finfo(np.float64).eps)        # avoid 0's in the matrix

        # Pre-allocation
        jm = np.zeros(shape=(max_iter, 1))
        g = np.zeros(shape=(n_clusters, data.shape[0]))


        for i in range(max_iter):
           
            # Perform one iteration of PFCM
            partition_matrix, typicality_matrix, centers, jm[i], g = self._pstepfcm(data=data, U=partition_matrix, T=typicality_matrix, m=m, a=a, b=b, n=n_clusters, g=g)
            
            # Stopping criterion: Stop if objective value does inrease < error
            if abs(jm[i] - jm[i-1]) < error:
                break

        return centers, partition_matrix, typicality_matrix, jm


    def _pstepfcm(self, data, U, T, g, m=2, n=2, a=0.5, b=0.5):
        #copy old u and t matrix and center locations
        um = U**m
        tf = T**n
        tfo = (1-T) ** n
        centers = (np.dot(a*um+b*tf, data).T/np.sum(a*um+b*tf, axis=1).T).T

        # Calculate distances between data points and cluster centers            
        dist = cdist(centers, data, metric='euclidean')
        dist = np.fmax(dist, np.finfo(np.float64).eps)

        #calculate value of objective funtion
        jm = np.sum(np.sum(np.power(dist, 2)*(a*um+b*tf), axis=0)) + np.sum(g*np.sum(tfo, axis=0))

        #calculate new u and t matrix
        g = um*np.power(dist, 2)/(np.sum(um, axis=0))
        tmp = np.power(dist, (-2/(m-1)))
        U = tmp/(np.sum(tmp, axis=0))
        tmpt = np.power((b/g)*np.power(dist, 2), (1/(n-1)))
        T = 1/(1+tmpt)

        return U, T, centers, jm, g
    
    ### Gustafson-Kessel

    def _gk(self, m=2, max_iter=1000, error=0.01):
        
        # Initialize the partition matrix
        u = np.random.dirichlet(np.ones(self.data.shape[0]), size=self.nr_clus)
        
        centers = []
        iteration = 0
        
        while iteration < max_iter:
            u_old = u.copy()   # keep old partition matrix to evaluate stopping criterium
            
            # Caluculate the locations of the cluster centers
            centers = self._next_centers_gk(data=self.data, u=u, m=m)
            
            # Calculate the covariance matrix
            f = self._covariance_gk(data=self.data, v=centers, u=u, m=m)
            
            # Calculate the distance between cluster centers and data points
            dist = self._distance_gk(data=self.data, v=centers, f=f)
            
            # calculate objective
            jm = (u * dist ** 2).sum()
            
            # Update the partition matrix
            u = self._next_u_gk(dist)
            iteration += 1

            # Stopping criteria
            if norm(u - u_old) < error:
                iteration = max_iter
                
        u=np.transpose(u)        
        return centers, u, jm 

    def _next_centers_gk(self, data, u, m=2):
        um = u ** m
        return ((um @ data).T / um.sum(axis=1)).T

    def _covariance_gk(self, data, v, u, m=2):
        um = u ** self.m

        denominator = um.sum(axis=1).reshape(-1, 1, 1)
        temp = np.expand_dims(data.reshape(data.shape[0], 1, -1) - v.reshape(1, v.shape[0], -1), axis=3)
        temp = np.matmul(temp, temp.transpose((0, 1, 3, 2)))
        numerator = um.transpose().reshape(um.shape[1], um.shape[0], 1, 1) * temp
        numerator = numerator.sum(0)

        return numerator / denominator

    def _distance_gk(self, data, v, f):
        dif = np.expand_dims(data.reshape(data.shape[0], 1, -1) - v.reshape(1, v.shape[0], -1), axis=3)
        determ = np.sign(np.linalg.det(f)) * (np.abs(np.linalg.det(f))) ** (1 / self.m)
        det_time_inv = determ.reshape(-1, 1, 1) * np.linalg.pinv(f)
        temp = np.matmul(dif.transpose((0, 1, 3, 2)), det_time_inv)
        output = np.matmul(temp, dif).squeeze().T
        return np.fmax(output, 1e-8)

    def _next_u_gk(self, d):
        power = float(1 / (self.m - 1))
        d = d.transpose()
        new_u = d.reshape((d.shape[0], 1, -1)).repeat(d.shape[-1], axis=1)
        new_u = np.power(d[:, None, :] / new_u.transpose((0, 2, 1)), power)
        new_u = 1 / new_u.sum(1)
        new_u = new_u.transpose()
        return new_u
    
    ### Relational Clustering
    
    def _RF_init_centers(self, number_of_clusters, number_of_data_points, relational_data, method='random_initialization'):
        # Initialize the cluster cenetrs for  relational fuzzy clustering
        if method == 'random_initialization':       # Use random numbers
            V = np.random.rand(number_of_clusters,number_of_data_points)
            V = V/V.sum(axis=1, keepdims=True)
        elif method == 'randomly_choose_c_rows':    # Use randomly selected data points as initial centers
            idx = np.random.choice(number_of_data_points,  size=number_of_clusters, replace=False)
            V = relational_data[idx,:]
            V = V/V.sum(axis=1, keepdims=True)
       
        return V


    def _rfcm(self, R, c, m = 2, epsilon = 0.005, maxIter = 1000, initType = 'random_initialization'):
     
        #   Relational Fuzzy c-Means (RFCM) for clustering dissimilarity data as
        #	proposed in [1]. RFCM is the relational dual of Fuzzy c-Means (FCM), 
        #	so it expects the input relational matrix R to be Euclidean. 
        #
        # Output:
        #               U: fuzzy partition matrix / membership matrix
        #               V: cluster centers/coefficients
        #               jm: Fitness function of the best solution
        #
        # Input:
        # R         - the relational (dissimilarity) data matrix of size n x n
        # c         - number of clusters to be identified
        # m         - fuzzifier, default 2
        # epsilon   - convergence criteria, default 0.0001
        # initType  - initialize relational cluster centers V, default random initialization
        #               random initialization
        #               randomly choose c rows from D
        # maxIter   - the maximum number fo iterations, default 100
        #
        # Refs:
        #   [1] Hathaway, R. J., Davenport, J. W., & Bezdek, J. C. (1989). Relational duals 
        #   of the c-means clustering algorithms. Pattern recognition, 22(2), 205-212. 
               
        # Initialize variables
        D = np.array(R)         # Relational data
        n=len(D)                # Number of data points
        d = np.zeros([c,n])
        numIter=0
        stepSize=epsilon
        
        # Initialize the membership matrix randomly
        U = np.random.rand(c, n)
        U = np.fmax(U, np.finfo(np.float64).eps)
        
        # Initialize the (relational) cluster centers
        V = self._RF_init_centers(number_of_clusters = c, number_of_data_points = n, relational_data = D, method=initType);
        
        # Begin the main loop:
        while numIter<maxIter and stepSize >= epsilon:
            U0 = U;
            
            # Compute the relational distances d between clusters centers V and data points D
            d=np.zeros([c,n])
            for i in range(0,c):
                Vi=V[i,:]
                tmp1 = D@Vi.T
                tmp2 = Vi@D@Vi.T/2
                d[i,:]= tmp1 - tmp2
                
            
            # Update the partition matrix U
            d = np.power(d,1/(m-1))
            tmp1 = np.divide(1,d)
            tmp2 = np.ones([c,1])*sum(tmp1)
            U = np.divide(np.divide(1,d),tmp2)
            print(U)
            
            # Update cluster centers V
            V = np.power(U,m)  
            V = V/V.sum(axis=1, keepdims=True)
            
            # Calculate the fitness value
            jm = (U * d ** 2).sum()
            
            # Update the step size
            stepSize=np.amax(abs(U-U0))

            if self._verbose == True:
                print('pyFUME just finished iteration '+ str(numIter) + ' of the RFCM clustering algoritm. The (maximum) difference between the previous and the newly computed partition matrix is' + str(stepSize) + '.')
            
            numIter = numIter + 1;
        
        # Return the cluster center location, the membership matrix and the fitness value
        return V, U, jm
        




from .pyfume import pyFUME


from .BuildTakagiSugeno import BuildTSFIS
from .Clustering import Clusterer
from .EstimateAntecendentSet import AntecedentEstimator
from .EstimateConsequentParameters import ConsequentEstimator
from .FeatureSelection import FeatureSelector
from .FireStrengthCalculator import FireStrengthCalculator
from .LoadData import DataLoader
from .Sampler import Sampler
from .simpfulfier import SimpfulConverter
from .SimpfulModelBuilder import SugenoFISBuilder
from .Splitter import DataSplitter
from .Tester import SugenoFISTester



import numpy as np

class DataNormalizer(object):
    """
        Creates an object that normalizes data.
        
        Args:
            data: The data that shpuld be normalized as a numpy array. 
    """
    
    def __init__(self, data):
        self.data=data
        
        
    def minmax(self):
        """
            Normalizes the data into a range between 0 and 1 using min-max normalization. 
    
            Returns:
                normalized_data: The normalized data.
        """
        normalized_data = (self.data - np.abs(self.data).min(axis=0)) / (np.abs(self.data).max(axis=0)-np.abs(self.data).min(axis=0))
        return normalized_data

    def zscore(self):
        """
            Normalizes the data using z-score normalization. 
    
            Returns:
                normalized_data: The normalized data.
        """
        normalized_data = (self.data - self.data.mean(axis=0)) / self.data.std(axis=0)
        return normalized_data


import numpy as np
from scipy.optimize import curve_fit
import numpy.matlib

class ConsequentEstimator(object):
    """
        Creates a new consequent estimator object.
        
        Args:
            x_train: The input data.
            y_train: The output data (true label/golden standard).
            firing_strengths: Matrix containing the degree to which each rule 
                fires for each data instance.
    """
    
    def __init__(self, x_train, y_train, firing_strengths):
        self.x_train=x_train
        self.y_train=y_train
        self.firing_strengths=firing_strengths
        
    def zero_order(self):
        """
            Estimates the consequent parameters of the zero-order Sugeno-Takagi model using normalized means.
        
            Args:
                df: default value returned when the sum of grades equals to one (default = 0).
        
            Returns:
                The parameters for the consequent function.
        """
        p=np.zeros((self.firing_strengths.shape[1]))
        for clus in range(0,self.firing_strengths.shape[1]): 
            fs=self.firing_strengths[:,clus]
            fs = np.fmax(fs, np.finfo(np.float64).eps)        # avoid 0's in the matrix
            normalized_weights=fs/fs.sum(0)
            s=np.multiply(normalized_weights, self.y_train) 
            p[clus]=sum(s)
        return p
            
    def suglms(self, global_fit=False, df=0):
        """
            Estimates the consequent parameters in the first-order Sugeno-Takagi model using least squares.
        
            Args:
                global_fit: Use the local (global_fit=False) or global (global_fit=True) least mean squares estimates. Global estimates functionality is still in beta mode, so use with caution.
                df: default value returned when the sum of grades equals to one (default = 0).
        
            Returns:
                The parameters for the consequent function.
        """
        
        x=self.x_train.copy()
        y=self.y_train.copy()
        f=self.firing_strengths.copy()
        
        # Check if input X contains one column of ones (for the constant). If not, add it.
        u=np.unique(x[:,-1])
        if u.shape[0]!=1 or u[0]!=1:
            x = np.hstack((x,np.ones((x.shape[0],1))))
    
        # Find the number of data points (mx & mf) , the number of variables (nx) and the
        # number of clusters (nf) 
        mx,nx=x.shape
        mf,nf=f.shape
        
        # Calculate the sum of the degree of fulfillement (DOF) for each data point
        sumDOF=np.sum(f, 1)
        
        
        # When degree of fulfillment is zero (which means no rule is applicable), set to one
        NoRule = sumDOF == 0
        sumDOF[NoRule] = 1
        sumDOF = np.matlib.repmat(sumDOF,nf,1).T
        
        
        if nf == 1:
           global_fit = False
        
        if global_fit == True:                # Global least mean squares estimates 

            # Still under construction!
            
            # Auxillary variables
            f1 = x.flatten()
            s = np.matlib.repmat(f1,nf,1).T
            xx = np.reshape(s, (nx,nf*mx), order='F')
            s = xx.T  
            x1=np.reshape(s,(mx,nf*nx),order='F') 
            x=x.T                                # reshape data matrix
            
            # (reshaped) vector of f devided by the sum of each row of f
            # (normalised membership degree)
            xx = (f.T.flatten()/sumDOF.T.flatten())
            
            # reshape partition matrix
            s= np.matlib.repmat(xx,nx,1).T
            f1 = np.reshape(s, (mx,nf*nx), order='F')                
            x1 = f1*x1
        
            # Find least squares solution
#            xp = np.linalg.lstsq(x1,y,rcond=None)
            
            # Perform QR decomposition
            Q,R = np.linalg.qr(x1) # qr decomposition of A
            Qy = np.dot(Q.T,y) # computing Q^T*b (project b onto the range of A)
            xx = np.linalg.solve(R,Qy)
            
            p=np.reshape(xx,(nf,nx), order='F')
        
            # Local models
            yl = np.transpose(x).dot(np.transpose(p))                                           #
            
            # Global model
            ym = x1.dot(p.flatten()) + df*NoRule
            ylm = yl.copy()
            
            # Mask all memberships < 0.2 with NaN's for plots
            ylm[f<0.2] = np.NaN
        
        elif global_fit == False:                                         # local weighted least mean squares estimates
            # Pre-allocate variable
            p=np.zeros([nf,nx])
            
            for i in range (0,nf):
                # Select firing strength of the selected rule
                w= f[:,i]
                
                # Weight input with firing strength
                xw = x * np.sqrt(w[:,np.newaxis])
                
                # Weight output with firing strength
                yw = y * np.sqrt(w)
                
                # Perform least squares with weighted input and output
                prm,_,_,_=np.linalg.lstsq(xw, yw, rcond=None)
                p[i]=prm
                       
        return p #,ym,yl,ylm



from .Splitter import DataSplitter
from .SimpfulModelBuilder import SugenoFISBuilder
from .Clustering import Clusterer
from .EstimateAntecendentSet import AntecedentEstimator
from .FireStrengthCalculator import FireStrengthCalculator
from .EstimateConsequentParameters import ConsequentEstimator
from .Tester import SugenoFISTester
import numpy as np

class FeatureSelector(object):
    """
        Creates a new feature selection object.
        
        Args:
            dataX: The input data.
            dataY: The output data (true label/golden standard).
            nr_clus: Number of clusters that should be identified in the data.
            variable_names: Names of the variables
            **kwargs: Additional arguments to change settings of the fuzzy model.
    """
    def __init__(self, dataX, dataY, nr_clus, variable_names, model_order='first', performance_metric='MAE', verbose=True, **kwargs):
        self.dataX=dataX
        self.dataY=dataY
        self.nr_clus = nr_clus
        self.variable_names = variable_names
        self.performance_metric = performance_metric
        self.model_order=model_order
        self.verbose=verbose
        
    def log_wrapper(self,**kwargs):
        """
            Performs feature selection using the wrapper method while also checking whether .
            
            Args:
                **kwargs: Additional arguments to change settings of the fuzzy model.
                
            Returns:
                Tuple containing (selected_features, selected_feature_names)
                    - selected_features: The indices of the selected features.
                    - selected_feature_names: The names of the selected features. 

        """
            
        # Create a training and validation set for the feature selection phase
        ds = DataSplitter()
        x_feat, y_feat, x_val, y_val = ds.holdout(self.dataX, self.dataY)
        
        # Set initial values for the performance
        old_performance=np.inf
        new_performance=np.inf
        perfs=[]
        
        # Create a set with the unselected (currently all) and selected (none yet) variables
        selected_features=[]
        unselected_features=list(range(0,np.size(x_feat,axis=1)))
        
        # Keep a (at this point empty) lists of which variables should be log-transformed
        logvars = []
        logvars_num = []
        temp_logvars_num = []
        temp_logvars  = []
        
        # Add a very small number before log transforming to avoid 0s in normalized data
        epsilon = np.finfo(np.float64).eps
        
        stop=False

        while stop == False: 
            perfs= [np.inf]*np.size(x_feat,axis=1)
            
            for f in [x for x in unselected_features if x != -1]:
                considered_features = selected_features + [f]
                var_names = [self.variable_names[i] for i in considered_features]
                
                # Prepare training sets
                feat=x_feat[:,considered_features]
                tmp = x_feat.copy()
                tmp[f] = np.log(tmp[f]+epsilon)
                log_feat=tmp[:,considered_features]

                
                # Prepare validation sets 
                x_validation=x_val[:,considered_features] 
                tmp=x_val.copy()
                tmp[f]=np.log(tmp[f]+epsilon)  
                log_x_validation=tmp[:,considered_features] 


                if self.verbose==True: 
                    print('Evaluating feature sub-set including:', var_names)
                
                try:
                    normal_perf = self._evaluate_feature_set(x_data=feat, y_data=y_feat, x_val=x_validation, y_val=y_val, nr_clus=self.nr_clus, var_names=var_names, model_order=self.model_order, performance_metric = self.performance_metric, **kwargs)
                    log_perf = self._evaluate_feature_set(x_data=log_feat, y_data=y_feat, x_val=log_x_validation, y_val=y_val, nr_clus=self.nr_clus, var_names=var_names, model_order=self.model_order, performance_metric = self.performance_metric, **kwargs)
                    if log_perf < normal_perf: 
                       temp_logvars.append(var_names[-1])
                       temp_logvars_num.append(f)
                       print(" * In this sub-set, the variable(s)", logvars + list([var_names[-1]]), "will be log-transformed.")
                    elif log_perf > normal_perf and len(logvars)>0:
                        print(" * In this sub-set, the variable(s)", logvars, "will be log-transformed.")

                    perfs[f] = min(normal_perf, log_perf)
                except RuntimeError:
                    raise Exception('ERROR: main module was not safely imported. Feature selection exploits multiprocessing, so please add a `if _name_ == `_main_`: `-line to your main script. See https://docs.python.org/2/library/multiprocessing.html#windows for further info.')
                    import sys
                    sys.exit(1)
                    
            new_performance=min(perfs)
            new_feature=unselected_features[perfs.index(new_performance)]
            
            del perfs
            
            if new_performance<old_performance: #*feature_selection_stop:
                selected_features.append(new_feature)
                                
                # Check if the last addded feature should be log-transformed and if so, perform the transformation
                if new_feature in temp_logvars_num:
                       logvars.append(self.variable_names[new_feature])
                       logvars_num.append(new_feature)
                    
                       x_feat[new_feature] = np.log(x_feat[new_feature]+epsilon)
                       x_val[new_feature] = np.log(x_val[new_feature]+epsilon)
                  
                unselected_features[new_feature]=-1
                old_performance=new_performance
                
                temp_logvars_num = []
                temp_logvars  = []

            else:
                print('***** FEATURE SELECTION ENDED *****')
                print('The selected features have a', self.performance_metric, 'of:', old_performance)
                stop = True 
       
        # Show the user which variables were selected, and which ones were log-transformed.
        selected_feature_names = [self.variable_names[i] for i in selected_features]
        print('The following features were selected:',  selected_feature_names)
        if len(logvars)>0:
            print('The following features were log-transformed:', logvars)
        elif len(logvars)==0:
            print('None of the selected features was log-transformed.')
            
        return selected_features, selected_feature_names, logvars_num, logvars

        
    def wrapper(self,**kwargs):
        """
            Performs feature selection using the wrapper method.
            
            Args:
                **kwargs: Additional arguments to change settings of the fuzzy model.
                
            Returns:
                Tuple containing (selected_features, selected_feature_names)
                    - selected_features: The indices of the selected features.
                    - selected_feature_names: The names of the selected features. 

        """
            
        # Create a training and validation set for the feature selection phase
        ds = DataSplitter()
        x_feat, y_feat, x_val, y_val = ds.holdout(self.dataX, self.dataY)
        
        # Set initial values for the performance
        old_performance=np.inf
        new_performance=np.inf
        perfs=[]
        
        # Create a set with the unselected (currently all) and selected (none yet) variables
        selected_features=[]
        unselected_features=list(range(0,np.size(x_feat,axis=1)))
        
        stop=False

        while stop == False: 
            perfs= [np.inf]*np.size(x_feat,axis=1)
            
            for f in [x for x in unselected_features if x != -1]:
                considered_features = selected_features + [f]
                var_names = [self.variable_names[i] for i in considered_features]
                feat=x_feat[:,considered_features]
                x_validation=x_val[:,considered_features] 
                if self.verbose==True: 
                    print('Evaluating feature sub set including:', var_names)
                
                try:
                    perfs[f] = self._evaluate_feature_set(x_data=feat, y_data=y_feat, x_val=x_validation, y_val=y_val, nr_clus=self.nr_clus, var_names=var_names, model_order=self.model_order, performance_metric = self.performance_metric, **kwargs)
                except RuntimeError:
                    raise Exception('ERROR: main module was not safely imported. Feature selection exploits multiprocessing, so please add a `if _name_ == `_main_`: `-line to your main script. See https://docs.python.org/2/library/multiprocessing.html#windows for further info.')
                    import sys
                    sys.exit(1)
                    
            new_performance=min(perfs)
            new_feature=unselected_features[perfs.index(new_performance)]
            
            #print('new feature', new_feature)
            del perfs
            
            if new_performance<old_performance: #*feature_selection_stop:
                selected_features.append(new_feature)
                unselected_features[new_feature]=-1
                old_performance=new_performance
            else:
                print('The selected features have a', self.performance_metric, 'of:', old_performance)
                stop = True 
       
        # selected_feature_names = self.variable_names[selected_features]
        selected_feature_names = [self.variable_names[i] for i in selected_features]
        print('The following features were selected:',  selected_feature_names)
        
        return selected_features, selected_feature_names

    def fst_pso_feature_selection(self,max_iter=100, min_clusters=2, max_clusters=10, performance_metric='MAE', **kwargs):
        """
            Perform feature selection using the FST-PSO [1] variant of the Integer and Categorical 
            PSO (ICPSO) proposed by Strasser and colleagues [2]. ICPSO hybridizes PSO and Estimation of Distribution 
            Algorithm (EDA), which makes it possible to convert a discrete problem to the (real-valued) 
            problem of estimating the distribution vector of a probabilistic model. Each fitness 
            evaluation a random solution is generated according to the probability distribution 
            encoded by the particle. Because the implementation is a variant on FST-PSO, the optimal 
            settings for the PSO are set automatically.

            If the number of clusters is set to None, this method simultaneously choses the optimal 
            number of clusters.

            [1] Nobile, M. S., Cazzaniga, P., Besozzi, D., Colombo, R., Mauri, G., & Pasi, G. (2018). 
            Fuzzy Self-Tuning PSO: A settings-free algorithm for global optimization. Swarm and 
            evolutionary computation, 39, 70-85.
            
            [2] Strasser, S., Goodman, R., Sheppard, J., & Butcher, S. (2016). A new discrete 
            particle swarm optimization algorithm. In Proceedings of the Genetic and Evolutionary 
            Computation Conference 2016 (pp. 53-60). 
            
            Args:
                max_iter: The maximum number of iterations used in the PSO (default = 10).
                min_clusters: The minimum number of clusters to be identified in the data set (only 
                when nr_clusters = None)
                max_clusters: The maximum number of clusters to be identified in the data set (only 
                when nr_clusters = None)
                performance_metric: The performance metric on which each solution is evaluated (default
                Mean Absolute Error (MAE))
                **kwargs: Additional arguments to change settings of the fuzzy model.
                
            Returns:
                Tuple containing (selected_features, selected_feature_names, optimal_number_clusters)
                    - selected_features: The indices of the selected features.
                    - selected_feature_names: The names of the selected features.
                    - optimal_number_clusters: If initially nr_clusters = None, this argument encodes the optimal number of clusters in the data set. If nr_clusters is not None, the optimal_number_clusters is set to nr_clusters.

        """
        
        from fstpso import FuzzyPSO


        FP = FuzzyPSO()
        
        # Create the search space for feature selection with number of dimensions D
        D = np.size(self.dataX,1)
        
        s=list([[True, False]]*D)
        
        # Add dimension for cluster number selection
        if self.nr_clus == None:
            s.append(list(range(min_clusters,max_clusters+1)))
        
        # Set search space
        FP.set_search_space_discrete(s)
        
        # Set the fitness function
        args={'x_train': self.dataX, 'y_train': self.dataY, 'verbose':self.verbose}
        FP.set_fitness(self._function, arguments=args)
        
        # solve problem with FST-PSO
        _, self.best_performance, self.best_solution = FP.solve_with_fstpso(max_iter=max_iter,verbose=False)
        
        if self.nr_clus == None:
            selected_features=self.best_solution[:-1]
        else:
           selected_features=self.best_solution 
    
        
        # Show best solution with fitness value
        varnams=[i for indx,i in enumerate(self.variable_names) if selected_features[indx]]
        print('The following features have been selected:', varnams, 'with a', self.performance_metric, 'of', round(self.best_performance,2))
        
        if self.nr_clus == None:
            optimal_number_clusters=self.best_solution[-1]
        else:
            optimal_number_clusters = self.nr_clus
            
            
        return selected_features, varnams, optimal_number_clusters

#    def fun(self, particle):
#        return sum(particle)
    
    def _function(self, particle, arguments, verbose=False, **kwargs):
        from itertools import compress 
        if self.nr_clus == None:
            A = arguments['x_train'][:,particle[:-1]]
            varnams=list(compress(self.variable_names, particle[:-1]))
            nr_clus=particle[-1]
        else:
            A = arguments['x_train'][:,particle[:]]
            varnams=list(compress(self.variable_names, particle[:]))
            nr_clus=self.nr_clus
        
        if A.shape[1]==0: ## If no features are selected, return a infinite high error
            error=np.inf
        else:
            error=self._evaluate_feature_set(x_data=A, y_data=arguments['y_train'], nr_clus=nr_clus, var_names=varnams, model_order=self.model_order, performance_metric=self.performance_metric, **kwargs)
            
        if verbose: print(" * Fitness: %.3f" % error)
        return error
    
    def _evaluate_feature_set(self, x_data, y_data, nr_clus, var_names, model_order='first', performance_metric='MAE', fs_number_of_folds=3, **kwargs):
        # Check settings and complete with default settings when needed
        if 'merge_threshold' not in kwargs.keys(): kwargs['merge_threshold'] = 1.0
        if 'cluster_method' not in kwargs.keys(): kwargs['cluster_method'] = 'fcm'        
        if kwargs['cluster_method'] == 'fcm':
            if 'fcm_m' not in kwargs.keys(): kwargs['fcm_m'] = 2
            if 'fcm_max_iter' not in kwargs.keys(): kwargs['fcm_maxiter'] = 1000
            if 'fcm_error' not in kwargs.keys(): kwargs['fcm_error'] = 0.005
        elif kwargs['cluster_method'] == 'fstpso':
            if 'fstpso_n_particles' not in kwargs.keys(): kwargs['fstpso_n_particles'] = None
            if 'fstpso_max_iter' not in kwargs.keys(): kwargs['fstpso_max_iter'] = 100
            if 'fstpso_path_fit_dump' not in kwargs.keys(): kwargs['fstpso_path_fit_dump'] = None
            if 'fstpso_path_sol_dump' not in kwargs.keys(): kwargs['fstpso_path_sol_dump'] = None
        if 'mf_shape' not in kwargs.keys(): kwargs['mf_shape'] = 'gauss'       
        if 'operators' not in kwargs.keys(): kwargs['operators'] = None
        if 'global_fit' not in kwargs.keys(): kwargs['global_fit'] = False  
        if 'verbose' not in kwargs.keys(): kwargs['verbose'] = False
                
        # Split the data using the hold-out method in a training (default: 75%) 
        # and test set (default: 25%).
        ds = DataSplitter()

        if fs_number_of_folds==1: ##### feauture selection with hold-out
            x_train, y_train, x_val, y_val = ds.holdout(x_data, y_data) 

            cl = Clusterer(x_train=x_train, y_train=y_train, nr_clus=nr_clus)               
                
            if kwargs['cluster_method'] == 'fcm':
                cluster_centers, partition_matrix, _ = cl.cluster(cluster_method='fcm', fcm_m=kwargs['fcm_m'], 
                    fcm_maxiter=kwargs['fcm_maxiter'], fcm_error=kwargs['fcm_error'])
            elif kwargs['cluster_method'] == 'fstpso':
                cluster_centers, partition_matrix, _ = cl.cluster(cluster_method='fstpso', 
                    fstpso_n_particles=kwargs['fstpso_n_particles'], fstpso_max_iter=kwargs['fstpso_max_iter'],
                    fstpso_path_fit_dump=kwargs['fstpso_path_fit_dump'], fstpso_path_sol_dump=kwargs['fstpso_path_sol_dump'])
            else:
                print('The requested clustering method is not (yet) implemented')
                 
            # Estimate the membership funtions of the system (default shape: gauss)
            antecedent_estimator = AntecedentEstimator(x_train, partition_matrix)

            antecedent_parameters = antecedent_estimator.determineMF(mf_shape=kwargs['mf_shape'], merge_threshold=kwargs['merge_threshold'])
            what_to_drop = antecedent_estimator._info_for_simplification

            # Build a first-order Takagi-Sugeno model using Simpful using dummy consequent parameters
            fsc=FireStrengthCalculator(antecedent_parameters, nr_clus, var_names, **kwargs)
            firing_strengths = fsc.calculate_fire_strength(x_train)
            
            # Estimate the parameters of the consequent
            ce = ConsequentEstimator(x_train, y_train, firing_strengths)
            
            if self.model_order=='first':
                consequent_parameters = ce.suglms()
            elif self.model_order== 'zero':
                consequent_parameters = ce.zero_order()

            # Build a first-order Takagi-Sugeno model using Simpful
            simpbuilder = SugenoFISBuilder(
                antecedent_parameters, 
                consequent_parameters, 
                var_names, 
                extreme_values = antecedent_estimator._extreme_values,
                operators=kwargs['operators'], 
                model_order=self.model_order,
                save_simpful_code=False, 
                fuzzy_sets_to_drop=what_to_drop,
                verbose=kwargs['verbose'])

            model = simpbuilder.simpfulmodel
            
            # Test the model
            test = SugenoFISTester(model=model, test_data=x_val, golden_standard=y_val,variable_names=var_names)
            performance= test.calculate_performance(metric=self.performance_metric)
        
        elif fs_number_of_folds>1:  ##### feauture selection with cross validation
            fold_indices = ds.kfold(data_length=np.shape(x_data)[0], number_of_folds=fs_number_of_folds)
            # perf=[np.inf]*fs_number_of_folds

            arg=[]
            for fold_number in range(0, fs_number_of_folds):
                
                # Choose the indices for training and testing for this fold
                tst_idx=fold_indices[fold_number]
                np.warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning)
                trn_idx=np.concatenate(np.delete(fold_indices, fold_number, axis=0))       # Use all indices, except the ones that are used for testing
                
                x_train = np.array([x_data[i,:] for i in trn_idx])
                x_val = np.array([x_data[i,:] for i in tst_idx])                      
                y_train = np.array([y_data[i] for i in trn_idx])
                y_val = np.array([y_data[i] for i in tst_idx]) 
                
                arg.append([x_train, y_train, x_val, y_val, nr_clus, var_names])
            
            try:
                from multiprocessing import Pool
            except ImportError:
                    raise Exception('pyFUME uses multiprocessing to parallelize computations, but couldn`t find \'multiprocessing\'. Please pip install multiprocessing to proceed.')
            
            try:
                with Pool(fs_number_of_folds) as p:
                    perf=p.starmap(func=self._create_model, iterable=arg)
            except RuntimeError:
                raise Exception('ERROR: main module was not safely imported. Feature selection exploits multiprocessing, so please add a `if _name_ == `_main_`: `-line to your main script. See https://docs.python.org/2/library/multiprocessing.html#windows for further info')
                
                # import os
                # import signal
                # os.kill(os.getppid(), signal.SIGTERM)
                
            performance = np.mean(perf)
            
            return performance


            #     # Cluster the training data (in input-output space)
            #     cl = Clusterer(x_train=x_train, y_train=y_train, nr_clus=nr_clus)                               
                
            #     if kwargs['cluster_method'] == 'fcm':
            #         cluster_centers, partition_matrix, _ = cl.cluster(cluster_method='fcm', fcm_m=kwargs['fcm_m'], 
            #             fcm_maxiter=kwargs['fcm_maxiter'], fcm_error=kwargs['fcm_error'])
            #     elif kwargs['cluster_method'] == 'gk':
            #         cluster_centers, partition_matrix, _ = cl.cluster(cluster_method='gk')
            #     elif kwargs['cluster_method'] == 'fstpso':
            #         cluster_centers, partition_matrix, _ = cl.cluster(cluster_method='fstpso', 
            #             fstpso_n_particles=kwargs['fstpso_n_particles'], fstpso_max_iter=kwargs['fstpso_max_iter'],
            #             fstpso_path_fit_dump=kwargs['fstpso_path_fit_dump'], fstpso_path_sol_dump=kwargs['fstpso_path_sol_dump'])
            #     else:
            #         print('The requested clustering method is not (yet) implemented')
                     
            #     # Estimate the membership funtions of the system (default shape: gauss)
            #     antecedent_estimator = AntecedentEstimator(x_train, partition_matrix)

            #     antecedent_parameters = antecedent_estimator.determineMF(mf_shape=kwargs['mf_shape'], merge_threshold=kwargs['merge_threshold'])
            #     what_to_drop = antecedent_estimator._info_for_simplification

            #     # Build a first-order Takagi-Sugeno model using Simpful using dummy consequent parameters
            #     fsc=FireStrengthCalculator(antecedent_parameters, nr_clus, var_names, **kwargs)
            #     firing_strengths = fsc.calculate_fire_strength(x_train)

            #     # Estimate the parameters of the consequent
            #     ce = ConsequentEstimator(x_train, y_train, firing_strengths)
            
            #     if self.model_order=='first':
            #         consequent_parameters = ce.suglms()
            #     elif self.model_order== 'zero':
            #         consequent_parameters = ce.zero_order()
                    
            #     # Build a first-order Takagi-Sugeno model using Simpful
            #     simpbuilder = SugenoFISBuilder(
            #         antecedent_parameters, 
            #         consequent_parameters, 
            #         var_names, 
            #         extreme_values = antecedent_estimator._extreme_values,
            #         operators=kwargs["operators"], 
            #         model_order=self.model_order,
            #         save_simpful_code=False, 
            #         fuzzy_sets_to_drop=what_to_drop,
            #         verbose=kwargs['verbose'])

            #     model = simpbuilder.simpfulmodel
                
            #     # Test the model
            #     test = SugenoFISTester(model=model, test_data=x_val, golden_standard=y_val,variable_names=var_names)
            #     perf[fold_number]= test.calculate_performance(metric=self.performance_metric) 

            # performance = np.mean(perf)

        # return performance
   
    def _create_model(self, x_train, y_train, x_test, y_test, nr_clus, var_names, **kwargs):
        # Cluster the training data (in input-output space)
        cl = Clusterer(x_train=x_train, y_train=y_train, nr_clus=nr_clus)
        cluster_centers, partition_matrix, _ = cl.cluster(cluster_method='fcm')                               
                
        # if kwargs['cluster_method'] == 'fcm':
        #     cluster_centers, partition_matrix, _ = cl.cluster(cluster_method='fcm', fcm_m=kwargs['fcm_m'], 
        #         fcm_maxiter=kwargs['fcm_maxiter'], fcm_error=kwargs['fcm_error'])
        # elif kwargs['cluster_method'] == 'gk':
        #     cluster_centers, partition_matrix, _ = cl.cluster(cluster_method='gk')
        # elif kwargs['cluster_method'] == 'fstpso':
        #     cluster_centers, partition_matrix, _ = cl.cluster(cluster_method='fstpso', 
        #         fstpso_n_particles=kwargs['fstpso_n_particles'], fstpso_max_iter=kwargs['fstpso_max_iter'],
        #         fstpso_path_fit_dump=kwargs['fstpso_path_fit_dump'], fstpso_path_sol_dump=kwargs['fstpso_path_sol_dump'])
        # else:
        #     print('The requested clustering method is not (yet) implemented')
             
        # # Estimate the membership funtions of the system (default shape: gauss)
        antecedent_estimator = AntecedentEstimator(x_train, partition_matrix)
        
        antecedent_parameters = antecedent_estimator.determineMF()
        #antecedent_parameters = antecedent_estimator.determineMF(mf_shape=kwargs['mf_shape'], merge_threshold=kwargs['merge_threshold'])
        what_to_drop = antecedent_estimator._info_for_simplification

        # Build a first-order Takagi-Sugeno model using Simpful using dummy consequent parameters
        fsc=FireStrengthCalculator(antecedent_parameters, nr_clus, var_names)
        firing_strengths = fsc.calculate_fire_strength(x_train)

        # Estimate the parameters of the consequent
        ce = ConsequentEstimator(x_train, y_train, firing_strengths)
    
        if self.model_order=='first':
            consequent_parameters = ce.suglms()
        elif self.model_order== 'zero':
            consequent_parameters = ce.zero_order()
            
        # Build a first-order Takagi-Sugeno model using Simpful
        simpbuilder = SugenoFISBuilder(
            antecedent_parameters, 
            consequent_parameters, 
            var_names, 
            extreme_values = antecedent_estimator._extreme_values, 
            model_order=self.model_order,
            save_simpful_code=False, 
            fuzzy_sets_to_drop=what_to_drop)
        
        # simpbuilder = SugenoFISBuilder(
        #     antecedent_parameters, 
        #     consequent_parameters, 
        #     var_names, 
        #     extreme_values = antecedent_estimator._extreme_values,
        #     operators=kwargs['operators'], 
        #     model_order=self.model_order,
        #     save_simpful_code=False, 
        #     fuzzy_sets_to_drop=what_to_drop,
        #     verbose=kwargs['verbose'])

        model = simpbuilder.simpfulmodel
        
        # Test the model
        test = SugenoFISTester(model=model, test_data=x_test, golden_standard=y_test, variable_names=var_names)
        perf = test.calculate_performance(metric=self.performance_metric) 
        return perf



import numpy as np

class Sampler(object):  
    """
        Creates a new Sampler object that makes it possible to oversample unbalanced data sets to make them more balanced.
        
        Args:
            train_x: The input data.
            train_y: The output data (true label/golden standard) on basis which will be sampled.
            number_of_bins: Number of clusters that should be identified in the data.
            histogram: True/False flag that determines whether a histogram of the frequencies of the output data will be plotted of both the old and new (= sampled) situation (default = False). The package 'matplotlib.pyplot' is required for this functionality.
    """ 

    def __init__(self,train_x, train_y, number_of_bins = 5, histogram = False):
        self.train_x = train_x
        self.train_y = train_y
        self.number_of_bins = number_of_bins
        self.histogram = histogram
        
    def oversample(self):
        """
        Created a more balanced data set by oversampling underrepresented data instances (based on values of the output variable) in the data set.

        Returns:
            Tuple containing (new_train_x, new_train_y)
                - new_train_x: The  oversampled input data metrix.
                - new_train_y: The oversampled output data matrix.
        """           
        print('Your training data will be oversampled.')
        data = np.append(self.train_x, self.train_y.reshape(len(self.train_y),1), axis=1)
        hist, bins = np.histogram(self.train_y, bins=self.number_of_bins)
        max_len=max(hist)
      
        for i in range(1,self.number_of_bins):
            if i==1:
                c = data[self.train_y <= bins[i]]
                indices = np.random.choice(len(c), max_len-len(c), replace=True)
                data_new= np.append(data, c[indices], axis=0)
            elif i>1 & i != self.number_of_bins-1:
                c = data[(self.train_y > bins[i-1]) & (self.train_y <= bins[i])]
                indices = np.random.choice(len(c), max_len-len(c), replace=True)
                data_new= np.append(data_new, c[indices], axis=0)
                        
        new_train_x = data_new[:,:-1]
        new_train_y = data_new[:,-1]
        
        if self.histogram == True:
            import matplotlib.pyplot as plt
            plt.hist(new_train_y, bins=self.number_of_bins, alpha=0.5, label='Oversampled data', color= 'blue')
            plt.hist(self.train_y, bins=self.number_of_bins, alpha=0.5, label='Raw data', color='red')
            plt.legend(loc='upper right')
            plt.show()
    
        return new_train_x, new_train_y


import simpful 

class SimpfulConverter(object):
    """    This object converts a description of a Fuzzy System into a readable
        Simpful project file.

        Args:
            input_variables_names: The names of the input variables of the fuzzy model.
            consequents_matrix: The parameters of the consequent function.
            fuzzy_sets: list containing a sub-list for each fuzzy set in the model. The sub-list should contain the shape of the fuzzy set (e.g. 'gauss') and the parameters of the function (e.g. mu, sigma). For example: ('gauss', 5, 1).
            model_order: The order of the fuzzy model ('zero' or 'first') (default = 'first').
            fuzzy_sets_to_drop: Fuzzy sets that should be disabled in the Simpful code, for example when flagged by GRABS for simplification (default = None).
            extreme_values: The values that should be used for the universe of discourse for the vaiables in the fuzzy model (default = None).
            operators: A list of strings, specifying fuzzy operators to be used instead of defaults (default = None). For more information see Simpful's documentation.
            verbose: Boolean (True/False) that indicates whether extra information will be printed in the user's console (default = True).

    """
    
    def __init__(self, 
            input_variables_names,
            consequents_matrix,
            fuzzy_sets,
            model_order = 'first',
            fuzzy_sets_to_drop = None,
            extreme_values = None,
            operators = None,
            verbose = False):
        super().__init__()
        self._input_variables = input_variables_names
        self._consequents_matrix = consequents_matrix
        self._clusters = len(self._consequents_matrix)
        self._fuzzy_sets = fuzzy_sets
        self._model_order = model_order
        self._fuzzy_sets_to_drop = fuzzy_sets_to_drop
        self._extreme_values = extreme_values
        self.verbose = verbose
        if self._fuzzy_sets_to_drop==None:
            self._fuzzy_sets_to_drop={}

        if self._model_order == 'first':  
            assert(len(self._input_variables)+1 == len(self._consequents_matrix[0]))
        
        if self.verbose: print(" * Detected %d rules / clusters" % self._clusters)

        self._source_code = []
        self._source_code.append( '# WARNING: this source code was automatically generated by pyFUME.' )
        self._source_code.append( "from simpful import *" )
        if operators is None:
            self._source_code.append("\nFS = FuzzySystem(show_banner=False)")
        else:
            # experimental, please test ASAP
            self._source_code.append("\nFS = FuzzySystem(operators="+str(operators)+")")

        
    def save_code(self, path):
        """
            Saves the Simpful code.
            
            Args:
                path: Path to the folder where the Simpful code should be saved.
        """
        code = self.generate_code()
        with open(path, "w") as fo:
            fo.write(code)
        if self.verbose == True:
            print (" * Code saved to file %s" % path)

    def generate_object(self):
        """
            Generates the executable object containing the fuzzy model.
        """
        code = self.generate_code()
        if self.verbose:
            exec(code, globals()) 
        elif self.verbose == False:
            import os
            import contextlib
            with open(os.devnull, "w") as f, contextlib.redirect_stdout(f):
                exec(code, globals()) 
        
        from copy import deepcopy
        self._fuzzyreasoner = deepcopy(FS)

    def generate_code(self):
        """
            Generates the Simpful code.
        """
        # rules
        rule_texts = self.create_rules()
        for i in range(1, self._clusters+1):
            self._source_code.append('RULE%d = "%s"' % (i, rule_texts[i-1]))
        self._source_code.append("FS.add_rules([%s])" % (", ". join(["RULE%d" % i for i in range(1,self._clusters+1)])))

        self._source_code.append("")

        # output functions
        
        if self._model_order == 'first':
            B = self._create_consequents()
            for i in range(self._clusters):
                self._source_code.append("FS.set_output_function('%s', '%s')" % ("fun%d" % (i+1),  B[i]))
        elif self._model_order == 'zero':
            for i in range(self._clusters):
                self._source_code.append("FS.set_crisp_output_value('%s', %s)" % ("fun%d" % (i+1),  self._consequents_matrix[i]))
        else:
            raise Exception("Model order not supported,"+self._model_order)
        self._source_code.append("")
    
        # fuzzy sets and membership functions
        result = self._create_fuzzy_sets()
        self._source_code.append(result)

        self._source_code.append("# end of automatically generated code #")

        return "\n".join(self._source_code)


    def _create_fuzzy_sets(self):
        
        j=0
        chunk = ""
        
        for num_var, var in enumerate(self._input_variables):

            subchunk = []
            for cluster in range(self._clusters):

                if (num_var, cluster) in self._fuzzy_sets_to_drop:
                    chunk+="# "

                #if self.verbose: print (" * Creating fuzzy set for variable %s, cluster%d" % (var, cluster+1))
                
                chunk += 'FS_%d = FuzzySet(' % (j+1)
                term = 'cluster%d' % (cluster+1)
                                
                fstype, params = self._fuzzy_sets[j]
                if fstype == 'gauss':
                    chunk += "function=Gaussian_MF(%f, %f), term='%s')" % (params[0], params[1], term) 
 
                elif fstype == 'gauss2':
                    chunk += "function=DoubleGaussian_MF(%f, %f, %f, %f), term='%s')" % (params[0], params[1], params[2], params[3], term) 

                elif fstype == 'sigmoid':
                    chunk += "function=Sigmoid_MF(%f, %f), term='%s')" % (params[0], params[1], term) 

                elif fstype == 'invgauss':
                    chunk += "function=InvGaussian_MF(%f, %f), term='%s')" % (params[0], params[1], term) 
                else:
                    raise Exception("Fuzzy set type not supported,"+fstype)
                if (num_var, cluster) not in self._fuzzy_sets_to_drop:
                    subchunk.append("FS_%d" % (j+1))
                #print ( self._fuzzy_sets[j] )
                j += 1
                chunk += "\n"
                # print(chunk)
            if self._extreme_values == None:
                chunk += "MF_%s = LinguisticVariable([%s], concept='%s')\n" % (var, ", ".join(subchunk), var )
            else:
                chunk += "MF_%s = LinguisticVariable([%s], concept='%s' , universe_of_discourse=%s)\n" % (var, ", ".join(subchunk), var, self._extreme_values[num_var] )
            chunk += "FS.add_linguistic_variable('%s', MF_%s)\n\n" % (var, var)

        return chunk

    def _create_consequents(self):
        result = []
        for row in self._consequents_matrix:
            result.append(("+".join(["%e*%s" % (value, name) for (name, value) in zip(self._input_variables, row[:-1])])))
            result[-1] += "+%e" % row[-1]
        return result

    def _create_antecedents(self):

        result = []

        for i in range(self._clusters):
            
            pieces = []
            for j, var in enumerate(self._input_variables):
                value = "cluster%d"% (i + 1)
                if (j,i) in self._fuzzy_sets_to_drop.keys():
                    value = "cluster%d" % (self._fuzzy_sets_to_drop[(j, i)] + 1)
                pieces.append("(%s IS %s)" % (var, value)) 
                
            chunk = (" AND ".join(pieces))
            result.append( chunk )

        return result
        
    def create_rules(self):
        A = self._create_antecedents()
        # B = self._create_consequents()
        B = ["fun%d" % (i+1) for i in range(self._clusters)]
        result = ["IF %s THEN (OUTPUT IS %s)" % (a,b) for a,b in zip(A,B)]
        return result


if __name__ == '__main__':
    
    SC = SimpfulConverter(
        input_variables_names = ["pippo", "pluto"],
        consequents_matrix = [[1,2,3], 
                              [2,3,5]],
        fuzzy_sets = [
                ["gauss", [0,1]],
                ["sigmoid", [1,2]],
                ["gauss2", [0,1,2,3]],
                ["invgauss", [0,1]]
                ]
    )
    
    SC.save_code("TEST.py")
    SC.generate_object()
    print(FS._mfs['pippo'])
    a


import numpy as np
from scipy.optimize import curve_fit

def is_complete(G):
    nodelist = G.nodes
    H = G.subgraph(nodelist)
    n = len(nodelist)
    return H.size() == n*(n-1)/2

class AntecedentEstimator(object):
    """
        Creates a new antecedent estimator object.
        
        Args:
            x_train: The input data.
            partition_matrix: The partition matrix of the input data (generated 
                by a clustering the data).
    """ 
    
    def __init__(self, x_train, partition_matrix):
        self.xtrain = x_train 
        self.partition_matrix = partition_matrix
        self._info_for_simplification = None
        self._calculate_all_extreme_values()
                        
        
    def determineMF(self, mf_shape='gauss', merge_threshold=1.0):
        """
            Estimates the parameters of the membership functions that are used 
            as antecedents of the fuzzy rules.
            
            Args:
                mf_shape: The desired shape of the fuzzy sets. The user can choose
                    from 'gauss' (gaussian), 'gauss2' (double gaussion) or 'sigmf' 
                    (sigmoidal) (default = gauss).
                merge_threshold: Threshold for the merging of fuzzy sets for 
                    the GRABS approach. By default no merging takes place 
                    (default = 1.0).
            Returns:
                A list with the estimated parameters of the membership functions (format if gauss: mu, sigma; if gauss2: mu1, sigma1, mu2, sigma2)
        """         
        mf_list=[]

        # mf_list is structured as follows:
        # - an outer list of V variables
        # - each item of the outer list contains C fuzzy set, one for each cluster
        number_of_variables = self.xtrain.shape[1]
        for i in range(0, number_of_variables):
            xin = self.xtrain[:,i]
            if all(y in (0, 1) for y in xin):           # Add noise to binary variables
                noise = np.random.normal(0,0.001,xin.shape[0]) 
                xin= xin + noise
                
            for j in range(0, self.partition_matrix.shape[1]):
                mfin = self.partition_matrix[:,j]
                mf, xx = self._convexMF(xin=xin, mfin=mfin)
                prm = self._fitMF(x=xx, mf=mf, mf_shape=mf_shape)
                mf_list.append(prm) 
        
        if merge_threshold < 1.0:
            self._check_similarities(mf_list, number_of_variables, threshold=merge_threshold)

        #print(self._info_for_simplification)

        return mf_list

    def is_subclique(self,G,nodelist):
        H = G.subgraph(nodelist)
        n = len(nodelist)
        return H.size() == n*(n-1)/2

    def _extreme_values_for_variable(self, v):
        return min(self.xtrain.T[v]), max(self.xtrain.T[v])

    def _calculate_all_extreme_values(self):
        num_variables = len(self.xtrain.T)
        self._extreme_values = [self._extreme_values_for_variable(v) for v in range(num_variables)]

    def _check_similarities(self, mf_list, number_of_variables,
            threshold=1., approx_points=100):

        number_of_clusters = len(mf_list)//number_of_variables
        
        from collections import defaultdict

        things_to_be_removed = defaultdict(list)


        """ 
            This function assesses the pair-wise similarities between 
            the clusters mapped on each variable.
            It returns a dictionary of this kind:
            variable -> list of similar couples for that variable + jaccard sim

        """

        for v in range(number_of_variables):

            mi, ma = self._extreme_values_for_variable(v)
            points = np.linspace(mi, ma, approx_points)

            for c1 in range(number_of_clusters):
                for c2 in range(c1+1, number_of_clusters):

                    index1 = v*number_of_clusters + c1
                    index2 = v*number_of_clusters + c2
                    funname1, params1 = mf_list[index1]
                    funname2, params2 = mf_list[index2]

                    if funname1== "gauss":
                        
                        from numpy import linspace, array

                        first_cluster = array([self._gaussmf(x, params1[0], params1[1]) for x in points])
                        second_cluster = array([self._gaussmf(x, params2[0], params2[1]) for x in points])

                        intersection = sum([min(x,y) for x,y in zip(first_cluster, second_cluster)])
                        union        = sum([max(x,y) for x,y in zip(first_cluster, second_cluster)])

                        jaccardsim = (intersection/union)

                        if jaccardsim>threshold:
                            things_to_be_removed[v].append([c1,c2,jaccardsim])

                            #print("%.2f is fine" % jaccardsim)

                    else:
                        raise Exception("Not implemented yet")

        #for k,v in things_to_be_removed.items():            print (k, v)
        #exit()
        """
            This function must return a dictionary of items like:
            (variable, cluster) -> mapped_cluster
        """

        self._info_for_simplification = {}
        for var_num, value in things_to_be_removed.items():

            subcomponents = self._create_graph(value)

            for subcomp in subcomponents:
                #print (is_complete(subcomp))

                if is_complete(subcomp):
                    retained = list(subcomp.nodes())[0]
                    #print ("retain: %d" % retained)
                    for el in list(subcomp.nodes()):
                        if el!=retained:
                            self._info_for_simplification[(var_num, el)]  = retained                    
                    
        dropped_stuff = self.get_number_of_dropped_fuzzy_sets()
        print (" * %d antecedent clauses will be simplified using a threshold %.2f" % (dropped_stuff, threshold))
        self._info_for_simplification

    def get_number_of_dropped_fuzzy_sets(self):
        return len(self._info_for_simplification)

    def _create_graph(self, list_of_arcs):
        from networkx import Graph, connected_components
        G = Graph()
        nodelist = []
        for arc in list_of_arcs:
            G.add_edge(arc[0], arc[1])
            nodelist.append(arc[0])
            nodelist.append(arc[1])
        S = [G.subgraph(c).copy() for c in connected_components(G)]
        return S
            
       
    def _convexMF(self, xin, mfin, norm=1, nc=1000):
        
        # Calculates the convex membership function that envelopes a given set of
        # data points and their corresponding membership values. 
        
        # Input:
        # Xin: N x 1 input domain (column vector)
        # MFin: N x 1 corresponding membership values 
        # nc: number of alpha cut values to consider (default=101)
        # norm: optional normalization flag (0: do not normalize, 1 : normalize, 
        # default=1)
        #
        # Output:
        # mf: membership values of convex function
        # x: output domain values    
        
        # Normalize the membership values (if requested)
        if norm == 1:
            mfin = np.divide(mfin, np.max(mfin))
        
        # Initialize auxilary variables
        acut = np.linspace(0,np.max(mfin),nc)
        mf= np.full(2*nc, np.nan)
        x=np.full(2*nc, np.nan)
        
        if np.any(mfin>0):
            x[0] = np.min(xin[mfin>0])
            x[nc]=np.max(xin[mfin>0])
            mf[0]=0
            mf[nc] = 0 
        
        # Determine the elements in the alpha cuts    
        for i in range(0,nc):
            if np.any(mfin>acut[i]):
                x[i]=np.min(xin[mfin>acut[i]])
                x[i+nc]=np.max(xin[mfin>acut[i]])
                mf[i]=acut[i]
                mf[i+nc]=acut[i]
                
        # # Determine the elements in the alpha cuts    
        # for i in range(0,nc):
        #     tmp1 = mfin>acut[i]
        #     if len(tmp1)==0:
        #         tmp=xin[tmp1]
        #         np.sort(tmp)
        #         x[i]=tmp[0]
        #         x[i+nc]=tmp[-1]
        #         mf[i]=acut[i]
        #         mf[i+nc]=acut[i]
                  
        #Delete NaNs
        idx=np.isnan(x)
        x=x[idx==False]
        mf=mf[idx==False]  
        
        # Sort vectors based on membership value (descending order)
        indmf=mf.argsort(axis=0)
        indmf=np.flipud(indmf)
        mf=mf[indmf]
        x=x[indmf]
        
        # Find duplicate values for x and onlykeep the ones with the highest membership value
        _, ind = np.unique(x, return_index=True, return_inverse=False, return_counts=False, axis=None)
        mf=mf[ind]
        x=x[ind]
        
        # Sort vectors based on x value (ascending order)
        indx=x.argsort(axis=0)
        mf=mf[indx]
        x=x[indx]
        
        xval=np.linspace(np.min(x),np.max(x),nc)
        mf=np.interp(xval, x, mf, left=None, right=None, period=None)
        x=xval;
        return mf, x
    
    def _fitMF(self,x,mf,mf_shape='gauss'):
        # Fits parametrized membership functions to a set of pointwise defined 
        # membership values.
        #
        # Input:
        # x:  N x 1 domain of input variable
        # mf: N x 1 membership values for input data x 
        # shape: Type of membership function to fit (possible values: 'gauss', 
        # 'gauss2' and 'sigmf')
        #
        # Output:
        # param: matrix of membership function parameters
    
        if mf_shape == 'gauss':
            # Determine initial parameters
            mu = sum(x * mf) / sum(mf)
            mf[mf==0] = np.finfo(np.float64).eps
            sig = np.mean(np.sqrt(-((x-mu)**2)/(2*np.log(mf))))
            
            # Fit parameters to the data using least squares
#            print('mu=', mu, 'sig=', sig)
            param, _ = curve_fit(self._gaussmf, x, mf, p0 = [mu, sig], bounds=((-np.inf, 0), (np.inf, np.inf)), maxfev = 10000)
       
        elif mf_shape == 'gauss2':
            # Determine initial parameters
            mu1 = x[mf>=0.95][0]
            mu2 = x[mf>=0.95][-1]
            xmf =x[mf>=0.5]
            sig1 = (mu1 - xmf[0])/(np.sqrt(2*np.log(2)))
            sig2 = (xmf[-1]-mu2)/(np.sqrt(2*np.log(2)))
            if sig1==0.0:
                sig1=0.1
            if sig2==0.0:
                sig2=0.1
            
            # Fit parameters to the data using least squares
#            print('mu1',mu1,'sig1',sig1,'mu2',mu2,'sig2',sig2)
            param, _ = curve_fit(self._gauss2mf, x, mf, p0 = [mu1, sig1, mu2, sig2], maxfev=1000, bounds=((-np.inf, 0,-np.inf, 0), (np.inf, np.inf,np.inf, np.inf)))
            
        elif mf_shape == 'sigmf':
            # Determine initial parameters
            if np.argmax(mf)-np.argmin(mf) > 0:         # if sloping to the right
                if len(x[mf>=0.5])>0:
                    c = x[mf>=0.5][0]
                    s = 1
                elif len(x[mf>=0.5])==0:                   # if there are no datapoints with membership larger than 0
                    c = x[0]
                    s = 1                    
            elif np.argmax(mf)-np.argmin(mf) < 0:       # if sloping to the left
                if len(x[mf<=0.5])>0:
                    c = x[mf<=0.5][0]
                    s = 1
                elif len(x[mf<=0.5])==0:                   # if there are no datapoints with membership smaller than 0
                    c = x[-1]
                    s = 1             
            # Fit parameters of the function to the data using non-linear least squares  
            try:
                param, _ = curve_fit(self._sigmf, x, mf, p0 = [c, s], maxfev=1000)
            except RuntimeError:
                print('pyFUME attempted to fit sigmoidal shaped membership functions, but was unable to find fitting parameters. pyFUME will now terminate. Please consider using a different shape for the membership functions.')
                import sys
                sys.exit()
        
        return mf_shape, param
        
    def _gaussmf(self,x, mu, sigma, a=1):
        # x:  (1D array)
        # mu: Center of the bell curve (float)
        # sigma: Width of the bell curve (float)
        # a: normalizes the bell curve, for normal fuzzy set a=1 (float) 
        return a * np.exp(-(x - mu)**2 / (2 * sigma**2))
    
    
    def _gauss2mf(self,x, mu1, sigma1, mu2, sigma2):
        # x: Data
        # mu1: Center of the leftside bell curve
        # sigma1: Standard deviation that determines the width of the leftside bell curve
        # mu2: Center of the rightside bell curve
        # sigma2: Standard deviation that determines the width of the rightside bell curve
        y = np.ones(len(x))
        idx1 = x <= mu1
        idx2 = x > mu2
        y[idx1] = self._gaussmf(x[idx1], mu1, sigma1)
        y[idx2] = self._gaussmf(x[idx2], mu2, sigma2)
        return y
    
    def _sigmf(self,x, c, s):
        # x: data
        # b: x where mf is 0.5
        # c: Controls 'width' of the sigmoidal region about `b` (magnitude); also
        #    which side of the function is open (sign). A positive value of `a`
        #    means the left side approaches 0.0 while the right side approaches 1,
        #    a negative value of `c` means the opposite.
        return 1. / (1. + np.exp(- s * (x - c)))


from collections import defaultdict
from math import sqrt
import numpy as np


class SugenoFISTester(object):
    """
    Creates a new Tester object to be able to calculate performance metrics of the fuzzy model.
    
    Args:
        model: The model for which the performance metrics should be calculated
        test_data: The data to be used to compute the performance metrics
        variable_names: A list of the variables names of the test data (which 
            should correspond with the variable names used in the model).
        golden_standard: The 'True' labels of the test data. If not provided, the 
            only predictions labels can be generated, but the error will not be 
            calculated (default = None).
        list_of_outputs: List of the output names (which should correspond with 
            the output names used in the model) (default: OUTPUT).
    """
    
    def __init__(self, model, test_data, variable_names, golden_standard=None, list_of_outputs=['OUTPUT']):
        super().__init__()
        self._model_to_test = model
        self._data_to_test = test_data
        self._golden_standard = golden_standard
        self._variable_names = variable_names
        self._list_of_outputs=list_of_outputs
        
    def predict(self):
        """
        Calculates the predictions labels of the test data using the fuzzy model.

        Returns:
            Tuple containing (result, error)
                - result: Prediction labels.
                - error: The difference between the prediction label and the 'true' label.
        """
        result = []
        for sample in self._data_to_test:
            for i, variable in enumerate(self._variable_names):
                self._model_to_test.set_variable(variable, sample[i])
            result.append(self._model_to_test.Sugeno_inference().get('OUTPUT'))
        result = np.array(result)
        if self._golden_standard is not  None:
            error = self._golden_standard - result
        else:
            error = np.nan
            # print('The true labels (golden standard) were not provided, so the error could not be calculated.')
        return result, error
    
    def calculate_performance(self, metric='MAE'):  
        """
        Calculates the performance of the model given the test data.

            Args:
                metric: The performance metric to be used to evaluate the model. Choose from: Mean Absolute Error 
                ('MAE'), Mean Squared Error ('MSE'),  Root Mean Squared Error ('RMSE'), Mean Absolute Percentage 
                Error ('MAPE').
        
        Returns:
            The performance as expressed by the chosen performance metric.
        """      
        if metric == 'MAE':
            err=self.calculate_MAE()
        elif metric == 'MSE':
            err=self.calculate_MSE()
        elif metric == 'RMSE':
            err=self.calculate_RMSE()        
        elif metric == 'MAPE':
            err=self.calculate_MAPE()
        else:
            print('The requested performance metric has not been implemented (yet).')
            
        return err
    
    def calculate_RMSE(self):
        """
        Calculates the Root Mean Squared Error of the model given the test data.
        
        Returns:
            The Root Mean Squared Error of the fuzzy model.
        """
        _, error=self.predict()
        return sqrt(np.mean(np.square(error)))
    
    
    def calculate_MSE(self):
        """
        Calculates the Mean Squared Error of the model given the test data.
        
        Returns:
            The Mean Squared Error of the fuzzy model.
        """
        _, error=self.predict()
        return np.mean(np.square(error))   
    
    def calculate_MAE(self):
        """
        Calculates the Mean Absolute Error of the model given the test data.
        
        Returns:
            The Mean Absolute Error of the fuzzy model.
        """
        _, error=self.predict()
        return np.mean(np.abs(error))
    
    def calculate_MAPE(self):
        """
        Calculates the Mean Absolute Percentage Error of the model given the test data.
        
        Returns:
            The Mean Absolute Percentage Error of the fuzzy model.
        """
        
        if self._golden_standard is None:
             raise Exception('To compute the MAPE the true label (golden standard) of the test data should be provided')
        
        _, error=self.predict()
        return np.mean(np.abs((error) / self._golden_standard)) * 100


import numpy as np
import random

class DataSplitter(object):
    """
        Creates an object that can (provide the indices to) split the data in a 
            training and test for model validation.
    """
    
#    def __init__(self):#, dataX, dataY):
#        self.data=np.loadtxt(datapath,delimiter=',')
#        self.dataX=self.data[:,0:-1]
#        self.dataY=self.data[:,-1]
#        self.dataX=dataX
#        self.dataY=dataY
        
    def holdout(self, dataX, dataY, percentage_training=0.75):
        """
            Splits the data in a training and test set using the hold-out method.
            
                Args:
                    dataX: The input data.
                    dataY: The output data (true label/golden standard).
                    percentage_training: Number between 0 and 1 that indicates the 
                        percentage of data that should be in the training data set 
                        (default = 0.75).
                    
                Returns:
                Tuple containing (x_train, y_train, x_test, y_test)
                        - x_train: Input variables of the training data.
                        - y_train: Output variables (true label/golden standard) of the training data.
                        - x_test: Input variables of the test data.
                        - y_test: Output variables (true label/golden standard) of the test data.
        """
        
        universe=set(range(0,np.shape(dataX)[0]))
        trn=np.random.choice(dataX.shape[0], int(round(percentage_training*dataX.shape[0])), replace=False)
        tst=list(universe-set(trn))

        x_train=dataX[trn]
        x_test=dataX[tst]
        y_train=dataY[trn]
        y_test=dataY[tst]
    
        return x_train, y_train, x_test, y_test

    def kfold(self, data_length, number_of_folds=10):
        """
            Provides the user with indices for 'k' number of  folds for the training 
                and testing of the model.
            
            Args:
                data_length: The total number of instances in the data sets 
                    (number of rows).
                number_of_folds: The number of folds the data should be split in 
                    (default = 10)
    
            Returns:
                A list with k (non-overlapping) sublists each containing the indices for one fold.
        """
        
        idx=np.arange(0, data_length)
        random.shuffle(idx)
         
        fold_indices= np.array_split(idx,number_of_folds)
        
        return fold_indices
    
        



from .simpfulfier import *
import sys

class SugenoFISBuilder(object):
    """
        Builds the executable Simpful model.
        
        Args:
            antecedent_sets: The parameters for the antecedent sets.
            consequent_parameters: The parameters for the consequent function.
            variable_names: The names of the variables.
            extreme_values: Extreme values to determine the universe of discourse. 
                If these are not set, the model will function but it will not 
                be possible to plot the membership functions (default = None).
            operators=None
            save_simpful_code: True/False, determines if the Simpful code will 
                be saved to the same folder as the script (default = True).
            fuzzy_sets_to_drop: Fuzzy sets that should be droppped from the 
                model (default = None).
    
    """
    def __init__(self, antecedent_sets, consequent_parameters, 
        variable_names, normalization_values = None, model_order = 'first', extreme_values=None, operators=None, 
        save_simpful_code=True, fuzzy_sets_to_drop=None, verbose=True):
        #super(SugenoFISBuilder, self).__init__()
        super().__init__()
        
        # if normalization_values != None:
        #     print('variable names:', variable_names)
        #     print('Antecedent sets:', antecedent_sets)
        #     print('Consequent parameters:', consequent_parameters)
        #     print('Normalization values:', normalization_values)
        #     sys.exit()            

       
        self._SC = SimpfulConverter(
            input_variables_names = variable_names,
            consequents_matrix = consequent_parameters,
            fuzzy_sets = antecedent_sets,
            operators = operators,
            extreme_values = extreme_values,
            model_order = model_order,
            fuzzy_sets_to_drop=fuzzy_sets_to_drop,
            verbose = verbose)
        
        # save_simpful_code can be either True (save on default file), 
        # False (not not save anything), or a path to a file.
        if save_simpful_code==True:
            self._SC.save_code("Simpful_code.py")
        elif save_simpful_code != False:
            self._SC.save_code(save_simpful_code)
       
        self._SC.generate_object()

        self.simpfulmodel = self._SC._fuzzyreasoner
    
    def get_model(self):
        return self.simpfulmodel


from .BuildTakagiSugeno import BuildTSFIS
from .LoadData import DataLoader
from .Splitter import DataSplitter
from .SimpfulModelBuilder import SugenoFISBuilder
from .Clustering import Clusterer
from .EstimateAntecendentSet import AntecedentEstimator
from .FireStrengthCalculator import FireStrengthCalculator
from .EstimateConsequentParameters import ConsequentEstimator
from .Tester import SugenoFISTester
from .FeatureSelection import FeatureSelector
from .Sampler import Sampler
from .simpfulfier import SimpfulConverter

import numpy as np

class pyFUME(object):
    """
        Creates a new fuzzy model.
        
        Args:
            datapath: The path to the csv file containing the input data (argument 'datapath' or 'dataframe' should be specified by the user).
            dataframe: Pandas dataframe containing the input data (argument 'datapath' or 'dataframe' should be specified by the user).
            nr_clus: Number of clusters that should be identified in the data (default = 2).
            process_categorical: Boolean to indicate whether categorical variables should be processed (default = False).
            method: At this moment, only Takagi Sugeno models are supported (default = 'Takagi-Sugeno')
            variable_names: Names of the variables, if not specified the names will be read from the first row of the csv file (default = None).
            merge_threshold: Threshold for GRABS to drop fuzzy sets from the model. If the jaccard similarity between two sets is higher than this threshold, the fuzzy set will be dropped from the model.
            **kwargs: Additional arguments to change settings of the fuzzy model.

        Returns:
            An object containing the fuzzy model, information about its setting (such as its antecedent and consequent parameters) and the different splits of the data.
    """
    def __init__(self, datapath=None, dataframe=None, nr_clus=2, process_categorical=False, method='Takagi-Sugeno', variable_names=None, merge_threshold=1., **kwargs):

        if datapath is None and dataframe is None:
            raise Exception("ERROR: a dataset was not specified. Please either use the datapath or dataframe arguments.")

         #if nr_clus<2 and nr_clus!=None:
         #    raise Exception("Number of clusters should be greater than 1.")

        self.datapath=datapath
        self.nr_clus=nr_clus
        self.method=method
        self.dropped_fuzzy_sets = 0
        #self.variable_names=variable_names

        if method=='Takagi-Sugeno' or method=='Sugeno':
            if datapath is not None:
                self.FIS = BuildTSFIS(datapath=self.datapath, nr_clus=self.nr_clus, variable_names=variable_names, process_categorical=process_categorical, merge_threshold=merge_threshold, **kwargs)
            else:
                self.FIS = BuildTSFIS(dataframe=dataframe, nr_clus=self.nr_clus, variable_names=variable_names, process_categorical=process_categorical, merge_threshold=merge_threshold, **kwargs)
            if merge_threshold < 1.0:
                self.dropped_fuzzy_sets = self.FIS._antecedent_estimator.get_number_of_dropped_fuzzy_sets()
        else:
            raise Exception ("This modeling technique has not yet been implemented.")

    def calculate_error(self, method="MAE"):
        """
        Calculates the performance of the model given the test data.

            Args:
                method: The performance metric to be used to evaluate the model (default = 'MAE'). Choose from: Mean Absolute Error 
                ('MAE'), Mean Squared Error ('MSE'),  Root Mean Squared Error ('RMSE'), Mean Absolute Percentage 
                Error ('MAPE').
        
        Returns:
            The performance as expressed by the chosen performance metric.
        """   
        if method=="MSE":
            return self._get_MSE()
        elif method=="MAE":
            return self._get_MAE()
        elif method=="MAPE":
            return self._get_MAPE()
        elif method=="RMSE":
            return self._get_RMSE()
        else:
            # return self._get_MSE()
            raise Exception("Method '%s' not implemented yet" % (method))
            
    def predict_test_data(self):
        """
        Calculates the predictions labels of the test data using the fuzzy model.

        Returns:
            Prediction labels.
        """
        #get the prediction for the test data set
        test = SugenoFISTester(model=self.FIS.model, test_data=self.FIS.x_test, variable_names=self.FIS.variable_names, golden_standard=self.FIS.y_test)
        pred, _ = test.predict()
        return pred
    
    def predict_label(self, xdata):
        """
        Calculates the predictions labels of a data set using the fuzzy model.

        Args:
            xdata: The input data (as numpy array with each row a different data instance and variables in the same order as in the original training data set) for which the labels should be calculated. 

        Returns:
            Prediction labels.
        """
        # Normalize the input data if needed
        if self.FIS.minmax_norm_flag==True:
            norm_val=self.FIS.normalization_values
            variable_names, min_values, max_values = zip(*norm_val)
            xdata = (xdata - np.array(min_values)) / (np.array(max_values) - np.array(min_values))
        
        #get the prediction for a new data set
        model = self.get_model()
        test = SugenoFISTester(model=model, test_data=xdata, golden_standard=None, variable_names=self.FIS.variable_names)
        pred, _ = test.predict()
        return pred
    
    def normalize_values(self, data):
        """
        Calculates the normalized values of a data point, using the same scaling 
        that was used to training data of the model. This method only works when 
        the data was normalized using the min-max method.

        Args:
            xdata: The input data (as numpy array with each row a different data instance and variables in the same order as in the original training data set) for which the normalized values should be calculated. 

        Returns:
            Normalized values.
        """        
        if self.FIS.minmax_norm_flag==True:
            norm_val=self.FIS.normalization_values
            variable_names, min_values, max_values = zip(*norm_val)
            normalized_data = (data - np.array(min_values)) / (np.array(max_values) - np.array(min_values)) 
           
        elif self.FIS.minmax_norm_flag==False:
            raise Exception('The model was not trained on normalized data, normalization is aborted.')    
        
        return normalized_data
    
    
    def denormalize_values(self, data):
        """
        Takes normalized data points, and returns the denormalized (raw) values
        of that data point. This method only works when during modeling the 
        data was normalized using the min-max method.

        Args:
            xdata: The input data (as numpy array with each row a different data instance and variables in the same order as in the original training data set) for which the normalized values should be calculated. 

        Returns:
            Normalized values.
        """        
        if self.FIS.minmax_norm_flag==True:
            if np.amin(data)<0 or np.amax(data) >1:
                print('WARNING: The given value(s) are not between 0 and 1, the denormalization is performed by extrapolating.')
            norm_val=self.FIS.normalization_values
            variable_names, min_values, max_values = zip(*norm_val)
            denormalized_data = (data * (np.array(max_values) - np.array(min_values))) + np.array(min_values)
        elif self.FIS.minmax_norm_flag==False:
            raise Exception('The model was not trained on normalized data, normalization is aborted.')    
        
        return denormalized_data

    def test_model(self, xdata, ydata, error_metric='MAE'):
        """
        Calculates the performance of the model using the given data.

        Args:
            xdata: The input data (as numpy array with each row a different data instance and variables in the same order as in the original training data set) for which the labels should be calculated. 
            ydata: The target data (as single-column numpy array).
            error_metric: The error metric in which the performance should be expressed (default = 'MAE'). Choose from: Mean Absolute Error ('MAE'), Mean Squared Error ('MSE'),  Root Mean Squared Error ('RMSE'), Mean Absolute Percentage Error ('MAPE').

        Returns:
            The performance as expressed in the chosen metric.
        """        
        #get the prediction for a new data set
        model = self.get_model()
        test = SugenoFISTester(model=model, test_data=xdata, golden_standard=ydata, variable_names=self.FIS.variable_names)
        metric= test.calculate_performance(metric=error_metric)
        return metric

    #######################
    ###     GETTERS     ###
    #######################

    def get_model(self):
        """
        Returns the fuzzy model created by pyFUME.

        Returns:
            The fuzzy model (as an executable object).
        """          
        if self.FIS.model is None:
            print ("ERROR: model was not created correctly, aborting.")
            exit(-1)
        else:
            return self.FIS.model
        
    def get_firing_strengths(self, data, normalize=True):
        """
        Calculates the (normalized) firing strength/ activition level of each rule for each data instance of the given data.

        Args:
            xdata: The input data (as numpy array with each row a different data instance and variables in the same order as in the original training data set) for which the labels should be calculated. 
            normalize: Boolean that indicates whether the retuned fiing strengths should be normalized (normalize = True) or not (normalize = False), When the firing strenghts are nomalized the summed fiing strengths for each data instance equals one.
        Returns:
            Firing strength/activition level of each rule (columns) for each data instance (rows).
        """          

        # Calculate the firing strengths
        fsc=FireStrengthCalculator(self.FIS.antecedent_parameters, self.FIS.nr_clus, self.FIS.variable_names)
        firing_strengths = fsc.calculate_fire_strength(data)
        if normalize == True:
            firing_strengths=firing_strengths/firing_strengths.sum(axis=1)[:,None]
        return firing_strengths
    
    def get_performance_per_fold(self):
        """
        Returns a list with the performances of each model that is created if crossvalidation is used when training..

        Returns:
            Perfomance of each cross-validation model.
        """
        return self.FIS.performance_metric_per_fold
    
    def get_fold_indices(self):
        """
        Returns a list with the fold indices of each model that is created if crossvalidation is used when training.

        Returns:
            Fold indices.
        """
        return self.FIS.fold_indices
        
    def _get_RMSE(self):
        # Calculate the mean squared error of the model using the test data set
        test = SugenoFISTester(model=self.FIS.model, test_data=self.FIS.x_test, golden_standard=self.FIS.y_test, variable_names=self.FIS.variable_names)
        RMSE = test.calculate_RMSE()
        #RMSE = list(RMSE.values())
        #print('The RMSE of the fuzzy system is', RMSE)
        return RMSE
    
    def _get_MSE(self):
        # Calculate the mean squared error of the model using the test data set
        test = SugenoFISTester(model=self.FIS.model, test_data=self.FIS.x_test, golden_standard=self.FIS.y_test, variable_names=self.FIS.variable_names)
        MSE = test.calculate_MSE()
        #RMSE = list(RMSE.values())
        #print('The RMSE of the fuzzy system is', RMSE)
        return MSE
        
    
    def _get_MAE(self):
        # Calculate the mean absolute error of the model using the test data set
        test = SugenoFISTester(model=self.FIS.model, test_data=self.FIS.x_test, variable_names=self.FIS.variable_names, golden_standard=self.FIS.y_test)
        MAE = test.calculate_MAE()
        return MAE
    
    def _get_MAPE(self):
        # Calculate the mean absolute percentage error of the model using the test data set
        test = SugenoFISTester(model=self.FIS.model, test_data=self.FIS.x_test, golden_standard=self.FIS.y_test, variable_names=self.FIS.variable_names)
        MAPE = test.calculate_MAPE()
        return MAPE
    
    def get_data(self, data_set='test'):
        """
        Returns the test or training data set.
        
        Args:
            data_set: Used to specify whether the function should return the training (data_set = "train"), test set (data_set = "test") or both training and test data (data_set = "all"). By default, the function returns the test set. 

        Returns:
            Tuple (x_data, y_data) containing the test or training data set.
        """  
      
        if data_set == 'train':
            return self.FIS.x_train, self.FIS.y_train
        elif data_set == 'test':
            return self.FIS.x_test, self.FIS.y_test
        elif data_set == 'all':
            xdata = np.concatenate((self.FIS.x_train, self.FIS.x_test), axis=0)
            ydata = np.concatenate((self.FIS.y_train, self.FIS.y_test), axis=0)
            return xdata, ydata
        else:
            print('Please specify whether you would like to receive the training (data_set = "train"), test set (data_set = "test") or all data (data_set = "all").')
            
    def get_cluster_centers(self):
        """
        Returns the cluster centers as identified by pyFUME.
        
        Returns:
            cluster centers.
        """  
        return self.FIS.cluster_centers 
    
    
    #######################
    ### PLOT FACILITIES ###
    #######################

    def plot_mf(self, variable_name, output_file='', highlight_element=None, highlight_mf=None, ax = None):
        """
        Uses Simpful's plotting facilities to plot the membership functions of
        the pyFUME model.

        Args:
            variable_name: The variable name whose membership functions should be plotted.
            output_file: Path and filename where the plot must be saved. By default, the file is not saved.
            highlight_element: Show the memberships of a specific element of the universe of discourse in the figure.
            highlight_mf: String indicating the linguistic term/fuzzy set to highlight in the plot.
            ax: The motplotlib ax where the variable will be plotted.

        """ 
        self.get_model().plot_variable(var_name = variable_name, outputfile = output_file, TGT = highlight_element, highlight = highlight_mf, ax = ax)
                
    def plot_all_mfs(self, output_file='', figures_per_row=4):
        """
        Plots the membership functions of all the variables  in the pyFUME model,
        each in their own sub figure.

        Args:
            output_file: path and filename where the plot must be saved.
            figures_per_row: The number of sub figures per row in the figure.
        """
        self.get_model().produce_figure(outputfile=output_file, max_figures_per_row=figures_per_row)
    
    def plot_consequent_parameters(self, rule_number, output_file=''):
        """
        Plots the consequent coeffecients of a given rule in a bar chart. If 
        the training data was normalized, the coeffiecients are plotted as-is. 
        If the data was not normalized, the coefficients are normalized to 
        enhance comparability.

        Args:
            output_file: path and filename where the plot must be saved.
            figures_per_row: The number of figures per row.
        """
        import matplotlib.pyplot as plt

        # Start counting from 1 instead of 0
        rule_number=rule_number-1
        
        # Get the required data from the pyFUME model
        labels= self.FIS.selected_variable_names
        consequent_parameters=self.FIS.consequent_parameters
        nr_rules=len(consequent_parameters)
        nr_variables = len(labels)
        
        # Standardize the parameters if data was not normalized
        if self.FIS.minmax_norm_flag == False:
            standard_deviations=np.std(self.FIS.x_train, axis=0)
            std_y=np.std(self.FIS.y_train)
            parameters=np.zeros((nr_variables,nr_rules))
            for rule in range(0,nr_rules):
                consequent=consequent_parameters[rule]
                n=np.zeros(nr_variables)
                for var in range(0,nr_variables):
                    std=standard_deviations[var]
                    parameter=consequent[var]
                    norm= (std/std_y)*parameter
                    n[var]=norm
                parameters[:,rule]= n
                
        # Used the raw values if data was normalized        
        elif self.FIS.minmax_norm_flag == True:
            parameters=np.zeros((nr_variables,nr_rules))
            for rule in range(0,nr_rules):
                consequent=consequent_parameters[rule]
                n=np.zeros(nr_variables)
                for var in range(0,nr_variables):
                    n[var]=consequent[var]
                parameters[:,rule]= n
        
        # Color the bars in the plot based on the relationship to the target variable
        cc=['colors']*len(parameters[:,rule_number])
        color_labels= ["Negatively related to target variable", "Positively related to target variable"]
        for n,val in enumerate(parameters[:,rule_number]):
            if val<0:
                cc[n]='firebrick'
            elif val>=0:
                cc[n]='navy' 
        
        # Create the plot
        fig, ax = plt.subplots(1,1)       
        h = plt.barh(labels, np.abs(parameters[:,rule_number]), align='center', color = cc)
        plt.grid(color='grey', linestyle='dotted', linewidth=1.5)
        plt.gca().invert_yaxis()
        if self.FIS.minmax_norm_flag == False:
            fig_title = 'Standardized consequent parameters for rule ' + str(rule_number+1)
        elif self.FIS.minmax_norm_flag == True:
            fig_title = 'Consequent parameters for rule ' + str(rule_number+1)    
        ax.set_title(fig_title)
        plt.legend(h, color_labels)
        fig.tight_layout()
        
        # Save the plot if requested, otherwise just show the plot to the user
        if output_file != "":
            fig.savefig(output_file)
        else:
            plt.show()

    def _denormalize_antecedent_set(self, data, normalization_values):
        """
        Takes a normalized antecedent set, and returns the denormalized parameters
        defining that set. This method only works when during modeling the 
        data was normalized using the min-max method.
    
        Args:
            xdata: The input data (as numpy array with each row a different data instance and variables in the same order as in the original training data set) for which the normalized values should be calculated. 
    
        Returns:
            Normalized values.
        """        
    
        (_, min_value, max_value) = normalization_values
        tmp = []
        for x in data[-1]:
            denormalized_data = (x * (np.array(max_value) - np.array(min_value))) + np.array(min_value)
            tmp.append(denormalized_data)
            
        denormalized_set = tuple([data[0], tmp])

        return denormalized_set
    
    def plot_denormalized_mf(self, variable_name, output_file='', highlight_element=None, highlight_mf=None, ax = None):
        normalization_values = self.FIS.normalization_values 
        antecedent_sets= self.FIS.antecedent_parameters
        
        # Check if variables were removed (during feature selection)
        to_keep=[]
        for i in range(0,len(normalization_values)):
            if normalization_values[i][0] in self.FIS.selected_variable_names:
                to_keep.append(i)
        
        # Keep only the values for variables that were selected
        normalization_values = [normalization_values[i] for i in to_keep]
        
        # Denormalize the antecedent set parameters
        denormed_antecedent_sets = []
        for i in range(0, len(antecedent_sets)):
            if i in range(0,len(self.FIS.selected_variable_names)):
                norm_vals=normalization_values[i]
            else:
                norm_vals=normalization_values[i-len(self.FIS.selected_variable_names)]
            denormed_set=self._denormalize_antecedent_set(antecedent_sets[i], norm_vals)
            denormed_antecedent_sets.append(denormed_set)
        
        UoD = []
        _, mi, ma = zip(*self.FIS.normalization_values)
        for i in range(0,len(mi)):
            # UoD.append(tuple((mi[i]-0.05*ma[i],ma[i]+0.05*ma[i])))
            UoD.append(tuple((mi[i],ma[i])))

        
        simpbuilder = SugenoFISBuilder(
            denormed_antecedent_sets, 
            np.tile(1, (self.nr_clus, len(self.FIS.selected_variable_names)+1)), 
            self.FIS.selected_variable_names, 
            extreme_values = UoD,
            save_simpful_code='trial.py', 
            fuzzy_sets_to_drop=self.FIS._antecedent_estimator._info_for_simplification,
            verbose = False)
        dummymodel = simpbuilder.simpfulmodel
        
        # Plot the requested variable using Simpful
        dummymodel.plot_variable(var_name = variable_name, outputfile = output_file, TGT = highlight_element, highlight = highlight_mf, ax = ax)

        


if __name__=='__main__':
    from numpy.random import seed
    seed(4)
   
    FIS = pyFUME(datapath='Concrete_data.csv', nr_clus=3, method='Takagi-Sugeno',
     merge_threshold=.8, operators=None)
    print ("The calculated error is:", FIS.calculate_error())

    FIS.get_model().produce_figure("bla.pdf")


from .SimpfulModelBuilder import SugenoFISBuilder

import numpy as np

class FireStrengthCalculator(object):
    """
        Creates a new fire strength calculator object.
        
        Args:
            antecedent_parameters: The parameters of the antecedent sets of the fuzzy model as given as output 
            by pyFUME's AntecedentEstimator clas (format: shape of the mf, parameters).
            nr_clus: Number of clusters in the data.
            variable_names: Names of the variables.
            fuzzy_sets_to_drop = Fuzzy sets identified by GRsABS to be dropped from the model (default = None).
            **kwargs: Additional arguments to change settings of the fuzzy model.
    """
    def __init__(self, antecedent_parameters, nr_clus, variable_names, fuzzy_sets_to_drop=None, **kwargs):

        self.antecedent_parameters = antecedent_parameters
        self.nr_clus = nr_clus
        self.variable_names = variable_names
        self.what_to_drop = fuzzy_sets_to_drop
        if 'operators' not in kwargs.keys(): kwargs['operators'] = None
        
        # Build a first-order Takagi-Sugeno model using Simpful using dummy consequent parameters
        simpbuilder = SugenoFISBuilder(
            self.antecedent_parameters, 
            np.tile(1, (self.nr_clus, len(self.variable_names)+1)), 
            self.variable_names, 
            extreme_values = None,
            operators=kwargs["operators"], 
            save_simpful_code=False, 
            fuzzy_sets_to_drop=self.what_to_drop,
            verbose = False)

        self.dummymodel = simpbuilder.simpfulmodel
        
    def calculate_fire_strength(self, data):
        """
            Calculates the firing strength per rule of the fuzzy model given a data set.
            
            Args:
                data: The data of which the firing strengths per rule should be calculated.
                
            Returns:
                The firing strengths per rule per data point (rows: data point index, column: rule/cluster number).
        """

        self.data=data
        
        # Create a dictionary to pass to Simpful
        input_dict = {}
        for i in range(len(self.variable_names)):
            input_dict[self.variable_names[i]] = self.data[:,i]
        
        # Calculate the firing strengths for each rule for each data point         
        self.firing_strengths = np.array(self.dummymodel.get_firing_strengths(input_values = input_dict))     
        
        # firing_strengths=[]
        # for i in range(0,len(self.data)):
        #     for j in range (0,len(self.variable_names)):
        #         self.dummymodel.set_variable(self.variable_names[j], self.data[i,j])
        #     firing_strengths.append(self.dummymodel.get_firing_strengths())
        # self.firing_strengths=np.array(firing_strengths)
        return self.firing_strengths


# from .Splitter import DataSplitter
from pyfume import Clusterer
import numpy as np
import scipy.spatial

class ClusterSelector(object):
    """
        Creates a new feature selection object.
        
        Args:
            dataX: The input data.
            dataY: The output data (true label/golden standard).
            nr_clus: Number of clusters that should be identified in the data.
            variable_names: Names of the variables
            **kwargs: Additional arguments to change settings of the fuzzy model.
    """
    def __init__(self, dataX, dataY, nr_clus, variable_names, **kwargs):
        self.dataX=dataX
        self.dataY=dataY
        self.nr_clus = nr_clus
        self.variable_names = variable_names
        
    def fst_pso_cluster_structure_selection(self,max_iter=100, min_clusters=2, max_clusters=10, performance_metric='MAE', **kwargs):
        """
            Perform feature selection using the FST-PSO [1] variant of the Integer and Categorical 
            PSO (ICPSO) proposed by Strasser and colleagues [2]. ICPSO hybridizes PSO and Estimation of Distribution 
            Algorithm (EDA), which makes it possible to convert a discrete problem to the (real-valued) 
            problem of estimating the distribution vector of a probabilistic model. Each fitness 
            evaluation a random solution is generated according to the probability distribution 
            encoded by the particle. Because the implementation is a variant on FST-PSO, the optimal 
            settings for the PSO are set automatically.

            If the number of clusters is set to None, this method simultaneously choses the optimal 
            number of clusters.

            [1] Nobile, M. S., Cazzaniga, P., Besozzi, D., Colombo, R., Mauri, G., & Pasi, G. (2018). 
            Fuzzy Self-Tuning PSO: A settings-free algorithm for global optimization. Swarm and 
            evolutionary computation, 39, 70-85.
            
            [2] Strasser, S., Goodman, R., Sheppard, J., & Butcher, S. (2016). A new discrete 
            particle swarm optimization algorithm. In Proceedings of the Genetic and Evolutionary 
            Computation Conference 2016 (pp. 53-60). 
            
            Args:
                max_iter: The maximum number of iterations used in the PSO (default = 10).
                min_clusters: The minimum number of clusters to be identified in the data set (only 
                when nr_clusters = None)
                max_clusters: The maximum number of clusters to be identified in the data set (only 
                when nr_clusters = None)
                performance_metric: The performance metric on which each solution is evaluated (default
                Mean Absolute Error (MAE))
                **kwargs: Additional arguments to change settings of the fuzzy model.
                
            Returns:
                Tuple containing (selected_features, selected_feature_names, optimal_number_clusters)
                    - selected_features: The indices of the selected features.
                    - selected_feature_names: The names of the selected features.
                    - optimal_number_clusters: If initially nr_clusters = None, this argument encodes the optimal number of clusters in the data set. If nr_clusters is not None, the optimal_number_clusters is set to nr_clusters.

        """
        
        from fstpso import FuzzyPSO


        FP = FuzzyPSO()
        
        # Create the search space for feature selection with number of dimensions D
        D = np.size(self.dataX,1)
        
        s=list([[True, False]]*D)
        
        # Add dimension for cluster number selection
        if self.nr_clus == None:
            s.append(list(range(min_clusters,max_clusters+1)))
        
        # Set search space
        FP.set_search_space_discrete(s)
        
        # Set the fitness function
        args={'x_train': self.dataX, 'y_train': self.dataY, 'verbose':self.verbose}
        FP.set_fitness(self._function, arguments=args)
        
        # solve problem with FST-PSO
        _, best_performance, best_solution = FP.solve_with_fstpso(max_iter=max_iter)
        
        if self.nr_clus == None:
            selected_features=best_solution[:-1]
        else:
           selected_features=best_solution 
    
        
        # Show best solution with fitness value
        varnams=[i for indx,i in enumerate(self.variable_names) if selected_features[indx]]
        print('The following features have been selected:', varnams, 'with a', self.performance_metric, 'of', round(best_performance,2))
        
        if self.nr_clus == None:
            optimal_number_clusters=best_solution[-1]
        else:
            optimal_number_clusters = self.nr_clus
            
            
        return selected_features, varnams, optimal_number_clusters

#    def fun(self, particle):
#        return sum(particle)
    
    def _function(self, particle, arguments, verbose=True, **kwargs):
        from itertools import compress 
        if self.nr_clus == None:
            A = arguments['x_train'][:,particle[:-1]]
            varnams=list(compress(self.variable_names, particle[:-1]))
            nr_clus=particle[-1]
        else:
            A = arguments['x_train'][:,particle[:]]
            varnams=list(compress(self.variable_names, particle[:]))
            nr_clus=self.nr_clus
        
        if A.shape[1]==0: ## If no features are selected, return a infinite high error
            error=np.inf
        else:
            error=self._evaluate_cluster_structure(x_data=A, y_data=arguments['y_train'], nr_clus=nr_clus, var_names=varnams, model_order=self.model_order, performance_metric=self.performance_metric, **kwargs)
            
        if verbose: print(" * Fitness: %.3f" % error)
        return error

        
    def _evaluate_cluster_structure(self, x_data, y_data, nr_clus, **kwargs):
        # Check settings and complete with default settings when needed
        if 'cluster_method' not in kwargs.keys(): kwargs['cluster_method'] = 'fcm'        
        if kwargs['cluster_method'] == 'fcm':
            if 'fcm_m' not in kwargs.keys(): kwargs['fcm_m'] = 2
            if 'fcm_max_iter' not in kwargs.keys(): kwargs['fcm_maxiter'] = 1000
            if 'fcm_error' not in kwargs.keys(): kwargs['fcm_error'] = 0.005
        elif kwargs['cluster_method'] == 'fstpso':
            if 'fstpso_n_particles' not in kwargs.keys(): kwargs['fstpso_n_particles'] = None
            if 'fstpso_max_iter' not in kwargs.keys(): kwargs['fstpso_max_iter'] = 100
            if 'fstpso_path_fit_dump' not in kwargs.keys(): kwargs['fstpso_path_fit_dump'] = None
            if 'fstpso_path_sol_dump' not in kwargs.keys(): kwargs['fstpso_path_sol_dump'] = None
                
        data = np.concatenate((x_data, y_data), axis=1)

        cl = Clusterer(data=data, nr_clus=nr_clus)               
            
        if kwargs['cluster_method'] == 'fcm':
            cluster_centers, partition_matrix, _ = cl.cluster(cluster_method='fcm', fcm_m=kwargs['fcm_m'], 
                fcm_maxiter=kwargs['fcm_maxiter'], fcm_error=kwargs['fcm_error'])
        elif kwargs['cluster_method'] == 'fstpso':
            cluster_centers, partition_matrix, _ = cl.cluster(cluster_method='fstpso', 
                fstpso_n_particles=kwargs['fstpso_n_particles'], fstpso_max_iter=kwargs['fstpso_max_iter'],
                fstpso_path_fit_dump=kwargs['fstpso_path_fit_dump'], fstpso_path_sol_dump=kwargs['fstpso_path_sol_dump'])
        else:
            print('The requested clustering method is not (yet) implemented')
            
        xb = self._xiebeni(data = data, centers = cluster_centers, um = partition_matrix, m=kwargs['fcm_m'])
            
        return xb

    def _xiebeni(data, centers, um, m=2):
    	n = data.shape[0]
    
    	dist = scipy.spatial.distance.cdist(data, centers, metric='sqeuclidean')
    	um_power = np.power(um,m)
    	numerator = np.sum(np.multiply(um_power,dist))
    
    	v2 = self._pairwise_squared_distances(centers,centers)
    	v2[v2 == 0.0] = np.inf
    
    	return numerator/(n*np.min(v2))
    
    def _pairwise_squared_distances(A, B):
        return scipy.spatial.distance.cdist(A, B)**2


from .LoadData import DataLoader
from .Splitter import DataSplitter
from .SimpfulModelBuilder import SugenoFISBuilder
from .Clustering import Clusterer
from .EstimateAntecendentSet import AntecedentEstimator
from .FireStrengthCalculator import FireStrengthCalculator
from .EstimateConsequentParameters import ConsequentEstimator
from .Tester import SugenoFISTester
from .FeatureSelection import FeatureSelector
from .Sampler import Sampler
import numpy as np


class BuildTSFIS(object):
    """
        Learns a a new  Takagi-Sugeno fuzzy model.
        
        Args:
            datapath: The path to the csv file containing the input data (argument 'datapath' or 'dataframe' should be specified by the user).
            dataframe: Pandas dataframe containing the input data (argument 'datapath' or 'dataframe' should be specified by the user).
            nr_clus: Number of clusters that should be identified in the data (default = 2).
            process_categorical: Boolean to indicate whether categorical variables should be processed (default = False).
            method: At this moment, only Takagi Sugeno models are supported (default = 'Takagi-Sugeno')
            variable_names: Names of the variables, if not specified the names will be read from the first row of the csv file (default = None).
            merge_threshold: Threshold for GRABS to drop fuzzy sets from the model. If the jaccard similarity between two sets is higher than this threshold, the fuzzy set will be dropped from the model.
            **kwargs: Additional arguments to change settings of the fuzzy model.

        Returns:
            An object containing the fuzzy model, information about its setting (such as its antecedent and consequent parameters) and the different splits of the data.
    """
    def __init__(self, datapath=None, dataframe=None, nr_clus=None, variable_names=None, 
            process_categorical=False, merge_threshold=1.0, **kwargs):

        self.datapath = datapath
        self.nr_clus = nr_clus
        self.variable_names = variable_names
        self._antecedent_estimator = None
        verbose = False

        if 'verbose' in kwargs.keys(): verbose = kwargs['verbose']
        # Check keyword-arguments and complete with default settings if necessary
        if 'model_order' not in kwargs.keys(): kwargs['model_order'] = 'first' 
        if 'normalize' not in kwargs.keys(): kwargs['normalize'] = False 
        if 'imputation' not in kwargs.keys(): kwargs['imputation'] = 'knn' # new
        if 'percentage_training' not in kwargs.keys(): kwargs['percentage_training'] = 0.75
        if 'oversampling' not in kwargs.keys(): kwargs['oversampling'] = False
        if kwargs['oversampling'] == True:
            if 'sampling_number_of_bins' not in kwargs.keys(): kwargs['sampling_number_of_bins'] = 2
            if 'sampling_histogram' not in kwargs.keys(): kwargs['sampling_histogram'] = False            
        if 'data_delimiter' not in kwargs.keys(): kwargs['data_delimiter'] = ','
        if 'data_split_method' not in kwargs.keys(): kwargs['data_split_method'] = 'hold-out'
        if 'feature_selection' not in kwargs.keys(): kwargs['feature_selection'] = None
        if 'fs_max_iter' not in kwargs.keys(): kwargs['fs_max_iter'] = 100
        if 'cluster_method' not in kwargs.keys(): kwargs['cluster_method'] = 'fcm'
        if 'm' not in kwargs.keys(): kwargs['m'] = '2'
        if kwargs['cluster_method'] == 'fcm':
            if 'fcm_max_iter' not in kwargs.keys(): kwargs['fcm_maxiter'] = 1000
            if 'fcm_error' not in kwargs.keys(): kwargs['fcm_error'] = 0.005
        elif kwargs['cluster_method'] == 'fst-pso':
            if 'fstpso_n_particles' not in kwargs.keys(): kwargs['fstpso_n_particles'] = None
            if 'fstpso_max_iter' not in kwargs.keys(): kwargs['fstpso_max_iter'] = 100
            if 'fstpso_path_fit_dump' not in kwargs.keys(): kwargs['fstpso_path_fit_dump'] = None
            if 'fstpso_path_sol_dump' not in kwargs.keys(): kwargs['fstpso_path_sol_dump'] = None
        elif kwargs['cluster_method'] == 'gk':
            if 'gk_max_iter' not in kwargs.keys(): kwargs['gk_maxiter'] = 1000
        if 'mf_shape' not in kwargs.keys(): kwargs['mf_shape'] = 'gauss'       
        if 'operators' not in kwargs.keys(): kwargs['operators'] = None
        if 'global_fit' not in kwargs.keys(): kwargs['global_fit'] = False  
        if 'save_simpful_code' not in kwargs.keys(): kwargs['save_simpful_code'] = True
        if 'verbose' not in kwargs.keys(): kwargs['verbose'] = True
        if 'cv_randomID' not in kwargs.keys(): kwargs['cv_randomID'] = False

        if self.nr_clus==None and kwargs['feature_selection'] == None:
            print('Error: please set pyFUME`s argument "nr_clus".')
            import sys
            sys.exit()
          

        # Load the data
        if self.datapath is None:
            dl=DataLoader(dataframe=dataframe, normalize=kwargs['normalize'], process_categorical=process_categorical, delimiter=kwargs['data_delimiter'], verbose=verbose)
        else:
            dl=DataLoader(self.datapath, normalize=kwargs['normalize'],  process_categorical=process_categorical, delimiter=kwargs['data_delimiter'], verbose=verbose)
        self.variable_names=dl.get_variable_names()
        
        if kwargs['normalize'] != False and kwargs['normalize'] != 'zscore':
            self.normalization_values=list(dl.get_normalization_values())
            self.minmax_norm_flag=True
        else:
            self.minmax_norm_flag = False
            self.normalization_values =  None
            
        self.dataX=dl.get_input_data()
        self.dataY=dl.get_target_data()
        
            
        # Create a DataSplitter object
        ds = DataSplitter()

        if kwargs['data_split_method'] == 'hold-out' or kwargs['data_split_method'] == 'holdout':
           
            if verbose: print(' * Hold-out method selected.')
            
            # Split the data using the hold-out method in a training (default: 75%) 
            # and test set (default: 25%).
            self.x_train, self.y_train, self.x_test, self.y_test = ds.holdout(dataX=self.dataX, dataY=self.dataY, percentage_training=kwargs['percentage_training'])
            # Check if there are any missing variables and impute them
            if np.isnan(self.dataX).any().any()== True:
                try:
                    from sklearn.impute import KNNImputer
                except ImportError:
                    raise Exception('pyFUME tried to impute missing values, but couldn`t find \'sklearn\'. Please pip install sklearn to proceed.')

                if verbose: print('Warning: Your data contains missing values that will be imputed using KNN.')   
                imputer = KNNImputer(n_neighbors=3, weights="uniform")
                self.x_train=imputer.fit_transform(self.x_train)
                self.x_test=imputer.fit_transform(self.x_test)

            
            if kwargs['oversampling'] == True:
                sample = Sampler(train_x = self.x_train, train_y=self.y_train, number_of_bins =  kwargs['sampling_number_of_bins'], histogram =  kwargs['sampling_histogram'])
                self.x_train, self.y_train = sample.oversample()
                    
            # Perform feature selection if requested
            if kwargs['feature_selection'] != None and kwargs['feature_selection'] != False:
                if 'performance_metric' not in kwargs.keys(): kwargs['performance_metric'] = 'MAE'
                fs = FeatureSelector(self.x_train, self.y_train, self.nr_clus, self.variable_names, model_order= kwargs['model_order'], performance_metric = kwargs['performance_metric'], verbose=kwargs['verbose'])
                
                # Keep copies of the training and test set before dropping unselected features
                self.x_train_before_fs=self.x_train.copy()
                self.x_test_before_fs=self.x_test.copy()
                            
                if kwargs['feature_selection'] == 'wrapper' or kwargs['feature_selection'] == 'sfs' or kwargs['feature_selection'] == 'SFS':
                    self.selected_feature_indices, self.variable_names=fs.wrapper()
                elif kwargs['feature_selection'] == 'logwrapper':
                    self.selected_feature_indices, self.selected_variable_names, self.log_indices, self.log_variable_names = fs.log_wrapper()
                elif kwargs['feature_selection'] == 'fst-pso' or kwargs['feature_selection'] == 'fstpso' or kwargs['feature_selection'] == 'pso' or kwargs['feature_selection'] == True:
                    self.selected_feature_indices, self.variable_names, self.nr_clus= fs.fst_pso_feature_selection(max_iter=kwargs['fs_max_iter']) 
                self.x_train = self.x_train[:, self.selected_feature_indices]
                self.x_test = self.x_test[:, self.selected_feature_indices]
                
            elif kwargs['feature_selection'] == None:
                self.selected_variable_names= self.variable_names
            
            # Cluster the data, log-transform when needed.
            if kwargs['feature_selection'] == 'logwrapper':
                # Use log transformed variables if needed
                self.log_x_train = self.x_train.copy()
                for i in self.log_indices:
                    self.log_x_train[i]= np.log(self.x_train[i])
                cl = Clusterer(x_train=self.log_x_train, y_train=self.y_train, nr_clus=self.nr_clus, verbose=verbose)
            else:                
                # Cluster the training data (in input-output space)
                cl = Clusterer(x_train=self.x_train, y_train=self.y_train, nr_clus=self.nr_clus, verbose=verbose)
            
            if kwargs['cluster_method'] == 'fcm':
                self.cluster_centers, self.partition_matrix, _ = cl.cluster(method='fcm', fcm_m=kwargs['m'], 
                    fcm_maxiter=kwargs['fcm_maxiter'], fcm_error=kwargs['fcm_error'])
            elif kwargs['cluster_method'] == 'gk':
                self.cluster_centers, self.partition_matrix, _ = cl.cluster(method='gk')
            elif kwargs['cluster_method'] == 'fst-pso':
                self.cluster_centers, self.partition_matrix, _ = cl.cluster(method='fstpso', 
                    fstpso_n_particles=kwargs['fstpso_n_particles'], fstpso_max_iter=kwargs['fstpso_max_iter'],
                    fstpso_path_fit_dump=kwargs['fstpso_path_fit_dump'], fstpso_path_sol_dump=kwargs['fstpso_path_sol_dump'])
            else: 
                print('ERROR: Choose a valid clustering method.')
                import sys
                sys.exit()
                    
                
            # Estimate the membership funtions of the system (default shape: gauss)
            self._antecedent_estimator = AntecedentEstimator(x_train=self.x_train, partition_matrix=self.partition_matrix)
    
            self.antecedent_parameters = self._antecedent_estimator.determineMF(mf_shape=kwargs['mf_shape'], merge_threshold=merge_threshold)
            what_to_drop = self._antecedent_estimator._info_for_simplification
            
            # Calculate the firing strengths
            fsc=FireStrengthCalculator(antecedent_parameters=self.antecedent_parameters, nr_clus=self.nr_clus, variable_names=self.variable_names,  **kwargs)
            self.firing_strengths = fsc.calculate_fire_strength(data=self.x_train)
  
            # Estimate the parameters of the consequent
            ce = ConsequentEstimator(x_train=self.x_train, y_train=self.y_train, firing_strengths=self.firing_strengths)
            if kwargs['model_order'] == 'first':
                self.consequent_parameters = ce.suglms()
            elif kwargs['model_order'] == 'zero':
                self.consequent_parameters = ce.zero_order()
            else:
                raise Exception("pyFUME currently only supports zero-order (model_order = 'zero') and first-order (model_order = 'first') fuzzy models.")
                    
            # Build a first-order Takagi-Sugeno model using Simpful
            simpbuilder = SugenoFISBuilder(
                self.antecedent_parameters, 
                self.consequent_parameters, 
                self.selected_variable_names, 
                normalization_values = self.normalization_values,
                extreme_values = self._antecedent_estimator._extreme_values,
                model_order=kwargs["model_order"],
                operators=kwargs["operators"], 
                save_simpful_code=kwargs['save_simpful_code'], 
                fuzzy_sets_to_drop=what_to_drop, verbose=verbose)
    
            self.model = simpbuilder.simpfulmodel
            
        elif kwargs['data_split_method']=='cross_validation' or kwargs['data_split_method']=='k-fold_cross_validation' or kwargs['data_split_method']=='crossvalidation' or kwargs['data_split_method']=='cv' or kwargs['data_split_method']=='kfold':
            if 'number_of_folds' not in kwargs.keys(): kwargs['number_of_folds'] = 10
            if verbose: print('K-fold cross validation was selected. The number of folds (k) equals', kwargs['number_of_folds'])
            if 'performance_metric' not in kwargs.keys(): kwargs['performance_metric'] = 'MAE'
            if 'save_kfold_models' not in kwargs.keys(): kwargs['save_kfold_models'] = 'True'

            #Create lists with test indices for each fold.
            self.fold_indices = ds.kfold(data_length=len(self.dataX), number_of_folds=kwargs['number_of_folds'])
            self.performance_metric = kwargs['performance_metric']
            # import indices from external file
            # import pickle
            # import pandas as pd

            #fold_indices=pd.read_csv('./fold_indices.csv', header=None)  
            #self.fold_indices=fold_indices.to_numpy()
                        
            # Create folder to store developed models
            if "save_kfold_models" == True:
                import os, datetime, pickle
    
                if kwargs['cv_randomID'] == True:
                    try:
                        import uuid
                    except ImportError:
                        raise Exception('pyFUME tried to generate random IDs, but couldn`t find \'uuid\'. Please pip install uuid to proceed.')
                    
                    self.folder_name= 'pyFUME runID ' + str(uuid.uuid4())
                elif kwargs['cv_randomID'] == False:
                    self.folder_name= 'pyFUME run ' + datetime.datetime.now().strftime("%Y-%m-%d %H.%M.%S")
                
                owd = os.getcwd()
                os.mkdir(self.folder_name)
                os.chdir('./' + self.folder_name)
                
                self.performance_metric_per_fold=[np.inf]*kwargs['number_of_folds']
            
            
            for fold_number in range(0, kwargs['number_of_folds']):
                if verbose: print('Training the model for fold', fold_number)
                tst_idx=self.fold_indices[fold_number]
                tst_idx = tst_idx[~np.isnan(tst_idx)]
                tst_idx = [int(x) for x in tst_idx]
                np.warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning)
                trn_idx=np.concatenate(np.delete(self.fold_indices, fold_number, axis=0))
                trn_idx = trn_idx[~np.isnan(trn_idx)]
                trn_idx = [int(x) for x in trn_idx]

                self.x_train = np.array([self.dataX[i,:] for i in trn_idx])
                self.x_test = np.array([self.dataX[i,:] for i in tst_idx])                      
                self.y_train = np.array([self.dataY[i] for i in trn_idx])
                self.y_test = np.array([self.dataY[i] for i in tst_idx]) 
                
                if np.isnan(self.dataX).any().any()== True:
                    try:
                        from sklearn.impute import KNNImputer
                    except ImportError:
                        raise Exception('pyFUME tried to impute missing values, but couldn`t find \'sklearn\'. Please pip install sklearn to proceed.')

                    if verbose: print('Warning: Your data contains missing values that will be imputed using KNN.')   
                    imputer = KNNImputer(n_neighbors=3, weights="uniform")
                    self.x_train=imputer.fit_transform(self.x_train)
                    self.x_test=imputer.fit_transform(self.x_test)
                
                if kwargs['oversampling'] == True:
                    sample = Sampler(train_x = self.x_train, train_y=self.y_train, number_of_bins =  kwargs['sampling_number_of_bins'], histogram =  kwargs['sampling_histogram'])
                    self.x_train, self.y_train = sample.oversample()
            
                # Perform feature selection if requested
                if kwargs['feature_selection'] != None and kwargs['feature_selection'] != False:
                    fs = FeatureSelector(self.x_train, self.y_train, self.nr_clus, self.variable_names, model_order= kwargs['model_order'], performance_metric = kwargs['performance_metric'], verbose=kwargs['verbose'])
                    self.x_train_before_fs=self.x_train.copy()
                    self.x_test_before_fs=self.x_test.copy()
                    
                    if kwargs['feature_selection'] == 'wrapper' or kwargs['feature_selection'] == 'sfs' or kwargs['feature_selection'] == 'SFS':
                        self.selected_feature_indices, self.selected_variable_names=fs.wrapper(feature_selection_stop=0.05)
                    elif kwargs['feature_selection'] == 'logwrapper':
                        self.selected_feature_indices, self.selected_variable_names, self.log_indices, self.log_variable_names = fs.log_wrapper()
                    elif kwargs['feature_selection'] == 'fst-pso' or kwargs['feature_selection'] == 'fstpso' or kwargs['feature_selection'] == 'pso' or kwargs['feature_selection'] == True:
                        self.selected_feature_indices, self.selected_variable_names, self.nr_clus= fs.fst_pso_feature_selection(max_iter=kwargs['fs_max_iter']) 
                    
                    self.x_train = self.x_train[:, self.selected_feature_indices]
                    self.x_test = self.x_test[:, self.selected_feature_indices]
                elif kwargs['feature_selection'] == None:
                    self.selected_variable_names= self.variable_names                    
                
                # Cluster the data, log-transform when needed.
                if kwargs['feature_selection'] == 'logwrapper':
                    # Use log transformed variables if needed
                    self.log_x_train = self.x_train.copy()
                    for i in self.log_indices:
                        self.log_x_train[i]= np.log(self.x_train[i])
                    cl = Clusterer(x_train=self.log_x_train, y_train=self.y_train, nr_clus=self.nr_clus, verbose=verbose)
                else:                
                    # Cluster the training data (in input-output space)
                    cl = Clusterer(x_train=self.x_train, y_train=self.y_train, nr_clus=self.nr_clus, verbose=verbose)
                
                
                if kwargs['cluster_method'] == 'fcm':
                    self.cluster_centers, self.partition_matrix, _ = cl.cluster(method='fcm', fcm_m=kwargs['m'], 
                        fcm_maxiter=kwargs['fcm_maxiter'], fcm_error=kwargs['fcm_error'])
                elif kwargs['cluster_method'] == 'fst-pso':
                    self.cluster_centers, self.partition_matrix, _ = cl.cluster(method='fstpso', 
                        fstpso_n_particles=kwargs['fstpso_n_particles'], fstpso_max_iter=kwargs['fstpso_max_iter'],
                        fstpso_path_fit_dump=kwargs['fstpso_path_fit_dump'], fstpso_path_sol_dump=kwargs['fstpso_path_sol_dump'])
                elif kwargs['cluster_method'] == 'gk':
                    self.cluster_centers, self.partition_matrix, _ = cl.cluster(method='gk')
                else: 
                    print('ERROR: Choose a valid clustering method.')
                    import sys
                    sys.exit()    
                
                # Estimate the membership funtions of the system (default shape: gauss)
                self._antecedent_estimator = AntecedentEstimator(self.x_train, self.partition_matrix)
        
                self.antecedent_parameters = self._antecedent_estimator.determineMF(mf_shape=kwargs['mf_shape'], merge_threshold=merge_threshold)
                what_to_drop = self._antecedent_estimator._info_for_simplification
        
                # Calculate the firing strengths
                fsc=FireStrengthCalculator(antecedent_parameters=self.antecedent_parameters, nr_clus=self.nr_clus, variable_names=self.selected_variable_names, **kwargs)
                self.firing_strengths = fsc.calculate_fire_strength(data=self.x_train)
                
                # Estimate the parameters of the consequent
                ce = ConsequentEstimator(self.x_train, self.y_train, self.firing_strengths)
                self.consequent_parameters = ce.suglms()
        
                # Build a first-order Takagi-Sugeno model using Simpful
                     
                simpbuilder = SugenoFISBuilder(
                    self.antecedent_parameters, 
                    self.consequent_parameters, 
                    self.selected_variable_names, 
                    normalization_values = self.normalization_values, 
                    extreme_values = self._antecedent_estimator._extreme_values,
                    operators=kwargs["operators"], 
                    save_simpful_code='Fold_' + str(fold_number) +'_Simpful_code.py', 
                    fuzzy_sets_to_drop=what_to_drop,
                    verbose=False)
        
                self.model = simpbuilder.simpfulmodel
                
                if "save_kfold_models" == True:
                    # Save the created model in the dedicated folder
                    filename = 'Fold_' + str(fold_number) + '.pickle'
                    pickle.dump(self, open(filename, 'wb'))
                
                tester=SugenoFISTester(model=self.model, test_data=self.x_test, variable_names=self.selected_variable_names, golden_standard=self.y_test)
                self.performance_metric_per_fold[fold_number]=tester.calculate_performance(metric=self.performance_metric)
                
            # set working directory back to where script is stored
            os.chdir(owd)
            print('The average ' + self.performance_metric + ' over ' + str(kwargs['number_of_folds']) +' folds is ' + str(np.mean(self.performance_metric_per_fold)) +' (with st. dev. ' + str(np.std(self.performance_metric_per_fold)) + '). \nThe best model was created in fold ' +  str(np.argmin(self.performance_metric_per_fold)) + ' with ' + self.performance_metric +  ' = ' + str(np.min(self.performance_metric_per_fold)) + '.')
                
        elif kwargs['data_split_method'] == 'no_split':
            if verbose: print('No test data will be split off, all data will be used for training.')
            
            self.x_train = self.dataX.copy()
            self.y_train = self.dataY.copy()
            
            if np.isnan(self.dataX).any().any()== True:
                try:
                    from sklearn.impute import KNNImputer
                except ImportError:
                    raise Exception('pyFUME tried to impute missing values, but couldn`t find \'sklearn\'. Please pip install sklearn to proceed.')

                if verbose: print('Warning: Your data contains missing values that will be imputed using KNN.')
    
                imputer = KNNImputer(n_neighbors=3, weights="uniform")
                self.x_train=imputer.fit_transform(self.x_train)
            
            if kwargs['oversampling'] == True:
                sample = Sampler(train_x = self.x_train, train_y=self.y_train, number_of_bins =  kwargs['sampling_number_of_bins'], histogram =  kwargs['sampling_histogram'])
                self.x_train, self.y_train = sample.oversample()
                    
            # Perform feature selection if requested
            if kwargs['feature_selection'] != None and kwargs['feature_selection'] != False:
                if 'performance_metric' not in kwargs.keys(): kwargs['performance_metric'] = 'MAE'
                fs=FeatureSelector(self.x_train, self.y_train, self.nr_clus, self.variable_names, model_order= kwargs['model_order'], performance_metric = kwargs['performance_metric'], verbose=kwargs['verbose'])
                
                # Keep copies of the training and test set before dropping unselected features
                self.x_train_before_fs=self.x_train.copy()
                            
                if kwargs['feature_selection'] == 'wrapper' or kwargs['feature_selection'] == 'sfs' or kwargs['feature_selection'] == 'SFS':
                    self.selected_feature_indices, self.variable_names=fs.wrapper()
                elif kwargs['feature_selection'] == 'logwrapper':
                    self.selected_feature_indices, self.selected_variable_names, self.log_indices, self.log_variable_names = fs.log_wrapper()
                elif kwargs['feature_selection'] == 'fst-pso' or kwargs['feature_selection'] == 'fstpso' or kwargs['feature_selection'] == 'pso' or kwargs['feature_selection'] == True:
                    self.selected_feature_indices, self.variable_names, self.nr_clus= fs.fst_pso_feature_selection(max_iter=kwargs['fs_max_iter']) 
                self.x_train = self.x_train[:, self.selected_feature_indices]
                
            elif kwargs['feature_selection'] == None:
                self.selected_variable_names= self.variable_names
                
            # Cluster the data, log-transform when needed.
            if kwargs['feature_selection'] == 'logwrapper':
                # Use log transformed variables if needed
                self.log_x_train = self.x_train.copy()
                for i in self.log_indices:
                    self.log_x_train[i]= np.log(self.x_train[i])
                cl = Clusterer(x_train=self.log_x_train, y_train=self.y_train, nr_clus=self.nr_clus, verbose=verbose)
            else:                
                # Cluster the training data (in input-output space)
                cl = Clusterer(x_train=self.x_train, y_train=self.y_train, nr_clus=self.nr_clus, verbose=verbose)
            
            if kwargs['cluster_method'] == 'fcm':
                self.cluster_centers, self.partition_matrix, _ = cl.cluster(method='fcm', fcm_m=kwargs['m'], 
                    fcm_maxiter=kwargs['fcm_maxiter'], fcm_error=kwargs['fcm_error'])
            elif kwargs['cluster_method'] == 'gk':
                self.cluster_centers, self.partition_matrix, _ = cl.cluster(method='gk')
            elif kwargs['cluster_method'] == 'fst-pso':
                self.cluster_centers, self.partition_matrix, _ = cl.cluster(method='fstpso', 
                    fstpso_n_particles=kwargs['fstpso_n_particles'], fstpso_max_iter=kwargs['fstpso_max_iter'],
                    fstpso_path_fit_dump=kwargs['fstpso_path_fit_dump'], fstpso_path_sol_dump=kwargs['fstpso_path_sol_dump'])
            else: 
                print('ERROR: Choose a valid clustering method.')
                import sys
                sys.exit()   
            
            # Estimate the membership funtions of the system (default shape: gauss)
            self._antecedent_estimator = AntecedentEstimator(self.x_train, self.partition_matrix)
    
            self.antecedent_parameters = self._antecedent_estimator.determineMF(mf_shape=kwargs['mf_shape'], merge_threshold=merge_threshold)
            what_to_drop = self._antecedent_estimator._info_for_simplification
            
            # Calculate the firing strengths
            fsc=FireStrengthCalculator(self.antecedent_parameters, self.nr_clus, self.variable_names, **kwargs)
            self.firing_strengths = fsc.calculate_fire_strength(self.x_train)
  
            # Estimate the parameters of the consequent
            ce = ConsequentEstimator(self.x_train, self.y_train, self.firing_strengths)
            if kwargs['model_order'] == 'first':
                self.consequent_parameters = ce.suglms()
            elif kwargs['model_order'] == 'zero':
                self.consequent_parameters = ce.zero_order()
            else:
                raise Exception("pyFUME currently only supports zero-order (model_order = 'zero') and first-order (model_order = 'first') fuzzy models.")
                    
            # Build a first-order Takagi-Sugeno model using Simpful
            simpbuilder = SugenoFISBuilder(
                self.antecedent_parameters, 
                self.consequent_parameters, 
                self.selected_variable_names,  
                normalization_values = self.normalization_values,
                extreme_values = self._antecedent_estimator._extreme_values,
                model_order=kwargs["model_order"],
                operators=kwargs["operators"], 
                save_simpful_code=kwargs['save_simpful_code'], 
                fuzzy_sets_to_drop=what_to_drop,
                verbose=verbose)
    
            self.model = simpbuilder.simpfulmodel
            
        else:
            print('ERROR: invalid data splitting method chosen. Training will be aborted.')
            import sys
            sys.exit()
